"""
Some old field names.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


_field_name_aliases = [
    ("GridLevel",                        "grid_level"),
    ("GridIndices",                      "grid_indices"),
    ("OnesOverDx",                       "ones_over_dx"),
    ("Ones",                             "ones"),
    #("CellsPerBin",                      "cells_per_bin"),
    ("SoundSpeed",                       "sound_speed"),
    ("RadialMachNumber",                 "radial_mach_number"),
    ("MachNumber",                       "mach_number"),
    ("CourantTimeStep",                  "courant_time_step"),
    #("ParticleVelocityMagnitude",        "particle_velocity_magnitude"),
    ("VelocityMagnitude",                "velocity_magnitude"),
    ("TangentialOverVelocityMagnitude",  "tangential_over_velocity_magnitude"),
    ("Pressure",                         "pressure"),
    ("Entropy",                          "entropy"),
    ("sph_r",                            "spherical_r"),
    ("sph_theta",                        "spherical_theta"),
    ("sph_phi",                          "spherical_phi"),
    ("cyl_R",                            "cylindrical_r"),
    ("cyl_z",                            "cylindrical_z"),
    ("cyl_theta",                        "cylindrical_theta"),
    ("cyl_RadialVelocity",               "cylindrical_radial_velocity"),
    ("cyl_RadialVelocityABS",            "cylindrical_radial_velocity_absolute"),
    ("cyl_TangentialVelocity",           "cylindrical_tangential_velocity"),
    ("cyl_TangentialVelocityABS",        "cylindrical_tangential_velocity_absolute"),
    ("DynamicalTime",                    "dynamical_time"),
    ("JeansMassMsun",                    "jeans_mass"),
    ("CellMass",                         "cell_mass"),
    ("TotalMass",                        "total_mass"),
    ("StarMassMsun",                     "star_mass"),
    ("Matter_Density",                   "matter_density"),
    ("ComovingDensity",                  "comoving_density"),
    ("Overdensity",                      "overdensity"),
    ("DensityPerturbation",              "density_perturbation"),
    ("Baryon_Overdensity",               "baryon_overdensity"),
    ("WeakLensingConvergence",           "weak_lensing_convergence"),
    ("CellVolume",                       "cell_volume"),
    ("ChandraEmissivity",                "chandra_emissivity"),
    ("XRayEmissivity",                   "xray_emissivity"),
    ("SZKinetic",                        "sz_kinetic"),
    ("SZY",                              "szy"),
    ("AveragedDensity",                  "averaged_density"),
    ("DivV",                             "div_v"),
    ("AbsDivV",                          "div_v_absolute"),
    ("Contours",                         "contours"),
    ("tempContours",                     "temp_contours"),
    ("SpecificAngularMomentumX",         "specific_angular_momentum_x"),
    ("SpecificAngularMomentumY",         "specific_angular_momentum_y"),
    ("SpecificAngularMomentumZ",         "specific_angular_momentum_z"),
    ("AngularMomentumX",                 "angular_momentum_x"),
    ("AngularMomentumY",                 "angular_momentum_y"),
    ("AngularMomentumZ",                 "angular_momentum_z"),
    #("ParticleSpecificAngularMomentumX", "particle_specific_angular_momentum_x"),
    #("ParticleSpecificAngularMomentumY", "particle_specific_angular_momentum_y"),
    #("ParticleSpecificAngularMomentumZ", "particle_specific_angular_momentum_z"),
    #("ParticleAngularMomentumX",         "particle_angular_momentum_x"),
    #("ParticleAngularMomentumY",         "particle_angular_momentum_y"),
    #("ParticleAngularMomentumZ",         "particle_angular_momentum_z"),
    #("ParticleRadius",                   "particle_radius"),
    ("Radius",                           "radius"),
    ("RadialVelocity",                   "radial_velocity"),
    ("RadialVelocityABS",                "radial_velocity_absolute"),
    ("TangentialVelocity",               "tangential_velocity"),
    ("CuttingPlaneVelocityX",            "cutting_plane_velocity_x"),
    ("CuttingPlaneVelocityY",            "cutting_plane_velocity_y"),
    ("CuttingPlaneBX",                   "cutting_plane_magnetic_field_x"),
    ("CuttingPlaneBy",                   "cutting_plane_magnetic_field_y"),
    ("MeanMolecularWeight",              "mean_molecular_weight"),
    ("particle_density",                 "particle_density"),
    ("ThermalEnergy",                    "thermal_energy"),
    ("TotalEnergy",                      "total_energy"),
    ("MagneticEnergy",                   "magnetic_energy"),
    ("GasEnergy",                        "thermal_energy"),
    ("Gas_Energy",                       "thermal_energy"),
    ("BMagnitude",                       "b_magnitude"),
    ("PlasmaBeta",                       "plasma_beta"),
    ("MagneticPressure",                 "magnetic_pressure"),
    ("BPoloidal",                        "b_poloidal"),
    ("BToroidal",                        "b_toroidal"),
    ("BRadial",                          "b_radial"),
    ("VorticitySquared",                 "vorticity_squared"),
    ("gradPressureX",                    "grad_pressure_x"),
    ("gradPressureY",                    "grad_pressure_y"),
    ("gradPressureZ",                    "grad_pressure_z"),
    ("gradPressureMagnitude",            "grad_pressure_magnitude"),
    ("gradDensityX",                     "grad_density_x"),
    ("gradDensityY",                     "grad_density_y"),
    ("gradDensityZ",                     "grad_density_z"),
    ("gradDensityMagnitude",             "grad_density_magnitude"),
    ("BaroclinicVorticityX",             "baroclinic_vorticity_x"),
    ("BaroclinicVorticityY",             "baroclinic_vorticity_y"),
    ("BaroclinicVorticityZ",             "baroclinic_vorticity_z"),
    ("BaroclinicVorticityMagnitude",     "baroclinic_vorticity_magnitude"),
    ("VorticityX",                       "vorticity_x"),
    ("VorticityY",                       "vorticity_y"),
    ("VorticityZ",                       "vorticity_z"),
    ("VorticityMagnitude",               "vorticity_magnitude"),
    ("VorticityStretchingX",             "vorticity_stretching_x"),
    ("VorticityStretchingY",             "vorticity_stretching_y"),
    ("VorticityStretchingZ",             "vorticity_stretching_z"),
    ("VorticityStretchingMagnitude",     "vorticity_stretching_magnitude"),
    ("VorticityGrowthX",                 "vorticity_growth_x"),
    ("VorticityGrowthY",                 "vorticity_growth_y"),
    ("VorticityGrowthZ",                 "vorticity_growth_z"),
    ("VorticityGrowthMagnitude",         "vorticity_growth_magnitude"),
    ("VorticityGrowthMagnitudeABS",      "vorticity_growth_magnitude_absolute"),
    ("VorticityGrowthTimescale",         "vorticity_growth_timescale"),
    ("VorticityRadPressureX",            "vorticity_radiation_pressure_x"),
    ("VorticityRadPressureY",            "vorticity_radiation_pressure_y"),
    ("VorticityRadPressureZ",            "vorticity_radiation_pressure_z"),
    ("VorticityRadPressureMagnitude",    "vorticity_radiation_pressure_magnitude"),
    ("VorticityRPGrowthX",               "vorticity_radiation_pressure_growth_x"),
    ("VorticityRPGrowthY",               "vorticity_radiation_pressure_growth_y"),
    ("VorticityRPGrowthZ",               "vorticity_radiation_pressure_growth_z"),
    ("VorticityRPGrowthMagnitude",       "vorticity_radiation_pressure_growth_magnitude"),
    ("VorticityRPGrowthTimescale",       "vorticity_radiation_pressure_growth_timescale"),
    ("DiskAngle",                        "theta"),
    ("Height",                           "height"),
]

_field_units_aliases = [
    ("cyl_RCode",                             "code_length"),
    ("HeightAU",                              "au"),
    ("cyl_RadialVelocityKMS",                 "km/s"),
    ("cyl_RadialVelocityKMSABS",              "km/s"),
    ("cyl_TangentialVelocityKMS",             "km/s"),
    ("cyl_TangentialVelocityKMSABS",          "km/s"),
    ("CellMassMsun",                          "msun"),
    ("CellMassCode",                          "code_mass"),
    ("TotalMassMsun",                         "msun"),
    ("CellVolumeCode",                        "code_length"),
    ("CellVolumeMpc",                         "Mpc**3"),
    ("ParticleSpecificAngularMomentumXKMSMPC","km/s/Mpc"),
    ("ParticleSpecificAngularMomentumYKMSMPC","km/s/Mpc"),
    ("ParticleSpecificAngularMomentumZKMSMPC","km/s/Mpc"),
    ("RadiusMpc",                             "Mpc"),
    ("ParticleRadiusMpc",                     "Mpc"),
    ("ParticleRadiuskpc",                     "kpc"),
    ("Radiuskpc",                             "kpc"),
    ("ParticleRadiuskpch",                    "kpc"),
    ("Radiuskpch",                            "kpc"),
    ("ParticleRadiuspc",                      "pc"),
    ("Radiuspc",                              "pc"),
    ("ParticleRadiusAU",                      "au"),
    ("RadiusAU",                              "au"),
    ("ParticleRadiusCode",                    "code_length"),
    ("RadiusCode",                            "code_length"),
    ("RadialVelocityKMS",                     "km/s"),
    ("RadialVelocityKMSABS",                  "km/s"),
    ("JeansMassMsun",                         "msun"),
]



"""
API for Sunrise Export code



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.funcs import issue_deprecation_warning

issue_deprecation_warning(
    "The sunrise_exporter module has been deprecated. This code has been "
    "moved to the yt attic (https://github.com/yt-project/yt_attic) and "
    "will be removed in a future release.")

from .sunrise_exporter import \
    export_to_sunrise



"""
Data structures for HaloCatalog frontend.




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.utilities.on_demand_imports import _h5py as h5py
import numpy as np
import glob

from .fields import \
    HaloCatalogFieldInfo

from yt.frontends.ytdata.data_structures import \
    SavedDataset
from yt.funcs import \
    parse_h5_attr
from yt.geometry.particle_geometry_handler import \
    ParticleIndex
from yt.data_objects.static_output import \
    ParticleFile

class HaloCatalogParticleIndex(ParticleIndex):
    def _setup_filenames(self):
        template = self.dataset.filename_template
        ndoms = self.dataset.file_count
        cls = self.dataset._file_class
        if ndoms > 1:
            self.data_files = \
              [cls(self.dataset, self.io, template % {'num':i}, i)
               for i in range(ndoms)]
        else:
            self.data_files = \
              [cls(self.dataset, self.io,
                   self.dataset.parameter_filename, 0)]

class HaloCatalogHDF5File(ParticleFile):
    def __init__(self, ds, io, filename, file_id):
        with h5py.File(filename, "r") as f:
            self.header = dict((field, parse_h5_attr(f, field)) \
                               for field in f.attrs.keys())

        super(HaloCatalogHDF5File, self).__init__(ds, io, filename, file_id)
    
class HaloCatalogDataset(SavedDataset):
    _index_class = ParticleIndex
    _file_class = HaloCatalogHDF5File
    _field_info_class = HaloCatalogFieldInfo
    _suffix = ".h5"
    _con_attrs = ("cosmological_simulation",
                  "current_time", "current_redshift",
                  "hubble_constant", "omega_matter", "omega_lambda",
                  "domain_left_edge", "domain_right_edge")

    def __init__(self, filename, dataset_type="halocatalog_hdf5",
                 n_ref = 16, over_refine_factor = 1, units_override=None,
                 unit_system="cgs"):
        self.n_ref = n_ref
        self.over_refine_factor = over_refine_factor
        super(HaloCatalogDataset, self).__init__(filename, dataset_type,
                                                 units_override=units_override,
                                                 unit_system=unit_system)

    def _parse_parameter_file(self):
        self.refine_by = 2
        self.dimensionality = 3
        nz = 1 << self.over_refine_factor
        self.domain_dimensions = np.ones(self.dimensionality, "int32") * nz
        self.periodicity = (True, True, True)
        prefix = ".".join(self.parameter_filename.rsplit(".", 2)[:-2])
        self.filename_template = "%s.%%(num)s%s" % (prefix, self._suffix)
        self.file_count = len(glob.glob(prefix + "*" + self._suffix))
        self.particle_types = ("halos")
        self.particle_types_raw = ("halos")
        super(HaloCatalogDataset, self)._parse_parameter_file()

    @classmethod
    def _is_valid(self, *args, **kwargs):
        if not args[0].endswith(".h5"): return False
        with h5py.File(args[0], "r") as f:
            if "data_type" in f.attrs and \
              parse_h5_attr(f, "data_type") == "halo_catalog":
                return True
        return False



"""
These are particle union objects.  These essentially alias one particle to
another, where the other can be one or several particle types.




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from .unions import Union

class ParticleUnion(Union):
    def __init__(self, name, sub_types):
        super(ParticleUnion, self).__init__(name, sub_types)



"""
Data structures for YTData frontend.




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2015, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from collections import \
    defaultdict
from numbers import \
    Number as numeric_type
import numpy as np
import os
import stat
import weakref

from .fields import \
    YTDataContainerFieldInfo, \
    YTGridFieldInfo

from yt.data_objects.grid_patch import \
    AMRGridPatch
from yt.data_objects.particle_unions import \
    ParticleUnion
from yt.data_objects.profiles import \
    Profile1DFromDataset, \
    Profile2DFromDataset, \
    Profile3DFromDataset
from yt.data_objects.static_output import \
    Dataset, \
    ParticleFile
from yt.extern.six import \
    string_types
from yt.funcs import \
    is_root, \
    parse_h5_attr
from yt.geometry.grid_geometry_handler import \
    GridIndex
from yt.geometry.particle_geometry_handler import \
    ParticleIndex
from yt.units import \
    dimensions
from yt.units.unit_registry import \
    UnitRegistry
from yt.units.yt_array import \
    uconcatenate, \
    YTQuantity
from yt.utilities.logger import \
    ytLogger as mylog
from yt.utilities.exceptions import \
    YTFieldTypeNotFound
from yt.utilities.on_demand_imports import \
    _h5py as h5py
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    parallel_root_only
from yt.utilities.tree_container import \
    TreeContainer
from yt.fields.field_exceptions import \
    NeedsGridType
from yt.data_objects.data_containers import \
    GenerationInProgress

_grid_data_containers = ["arbitrary_grid",
                         "covering_grid",
                         "smoothed_covering_grid"]

class SavedDataset(Dataset):
    """
    Base dataset class for products of calling save_as_dataset.
    """
    _con_attrs = ()

    def _parse_parameter_file(self):
        self.refine_by = 2
        with h5py.File(self.parameter_filename, "r") as f:
            for key in f.attrs.keys():
                v = parse_h5_attr(f, key)
                if key == "con_args":
                    v = v.astype("str")
                self.parameters[key] = v
            self._with_parameter_file_open(f)

        # if saved, restore unit registry from the json string
        if "unit_registry_json" in self.parameters:
            self.unit_registry = UnitRegistry.from_json(
                self.parameters["unit_registry_json"])
            # reset self.arr and self.quan to use new unit_registry
            self._arr = None
            self._quan = None
            for dim in ["length", "mass", "pressure",
                        "temperature", "time", "velocity"]:
                cu = "code_" + dim
                if cu not in self.unit_registry:
                    self.unit_registry.add(
                        cu, 1.0, getattr(dimensions, dim))
            if "code_magnetic" not in self.unit_registry:
                self.unit_registry.add("code_magnetic", 1.0,
                                       dimensions.magnetic_field)

        # if saved, set unit system
        if "unit_system_name" in self.parameters:
            unit_system = self.parameters["unit_system_name"]
            del self.parameters["unit_system_name"]
        else:
            unit_system = "cgs"
        # reset unit system since we may have a new unit registry
        self._assign_unit_system(unit_system)

        # assign units to parameters that have associated unit string
        del_pars = []
        for par in self.parameters:
            ustr = "%s_units" % par
            if ustr in self.parameters:
                if isinstance(self.parameters[par], np.ndarray):
                    to_u = self.arr
                else:
                    to_u = self.quan
                self.parameters[par] = to_u(
                    self.parameters[par], self.parameters[ustr])
                del_pars.append(ustr)
        for par in del_pars:
            del self.parameters[par]

        for attr in self._con_attrs:
            setattr(self, attr, self.parameters.get(attr))
        self.unique_identifier = \
          int(os.stat(self.parameter_filename)[stat.ST_CTIME])

    def _with_parameter_file_open(self, f):
        # This allows subclasses to access the parameter file
        # while it's still open to get additional information.
        pass

    def set_units(self):
        if "unit_registry_json" in self.parameters:
            self._set_code_unit_attributes()
            del self.parameters["unit_registry_json"]
        else:
            super(SavedDataset, self).set_units()

    def _set_code_unit_attributes(self):
        attrs = ('length_unit', 'mass_unit', 'time_unit',
                 'velocity_unit', 'magnetic_unit')
        cgs_units = ('cm', 'g', 's', 'cm/s', 'gauss')
        base_units = np.ones(len(attrs))
        for unit, attr, cgs_unit in zip(base_units, attrs, cgs_units):
            if attr in self.parameters and \
              isinstance(self.parameters[attr], YTQuantity):
                uq = self.parameters[attr]
            elif attr in self.parameters and \
              "%s_units" % attr in self.parameters:
                uq = self.quan(self.parameters[attr],
                               self.parameters["%s_units" % attr])
                del self.parameters[attr]
                del self.parameters["%s_units" % attr]
            elif isinstance(unit, string_types):
                uq = self.quan(1.0, unit)
            elif isinstance(unit, numeric_type):
                uq = self.quan(unit, cgs_unit)
            elif isinstance(unit, YTQuantity):
                uq = unit
            elif isinstance(unit, tuple):
                uq = self.quan(unit[0], unit[1])
            else:
                raise RuntimeError("%s (%s) is invalid." % (attr, unit))
            setattr(self, attr, uq)

class YTDataset(SavedDataset):
    """Base dataset class for all ytdata datasets."""

    _con_attrs = ("cosmological_simulation", "current_time",
                  "current_redshift", "hubble_constant",
                  "omega_matter", "omega_lambda",
                  "dimensionality", "domain_dimensions",
                  "periodicity",
                  "domain_left_edge", "domain_right_edge",
                  "container_type", "data_type")

    def _with_parameter_file_open(self, f):
        self.num_particles = \
          dict([(group, parse_h5_attr(f[group], "num_elements"))
                for group in f if group != self.default_fluid_type])

    def create_field_info(self):
        self.field_dependencies = {}
        self.derived_field_list = []
        self.filtered_particle_types = []
        self.field_info = self._field_info_class(self, self.field_list)
        self.coordinates.setup_fields(self.field_info)
        self.field_info.setup_fluid_fields()
        for ptype in self.particle_types:
            self.field_info.setup_particle_fields(ptype)

        self._setup_gas_alias()
        self.field_info.setup_fluid_index_fields()

        if "all" not in self.particle_types:
            mylog.debug("Creating Particle Union 'all'")
            pu = ParticleUnion("all", list(self.particle_types_raw))
            self.add_particle_union(pu)
        self.field_info.setup_extra_union_fields()
        mylog.debug("Loading field plugins.")
        self.field_info.load_all_plugins()
        deps, unloaded = self.field_info.check_derived_fields()
        self.field_dependencies.update(deps)

    def _setup_gas_alias(self):
        pass

    def _setup_override_fields(self):
        pass

class YTDataHDF5File(ParticleFile):
    def __init__(self, ds, io, filename, file_id):
        with h5py.File(filename, "r") as f:
            self.header = dict((field, parse_h5_attr(f, field)) \
                               for field in f.attrs.keys())

        super(YTDataHDF5File, self).__init__(ds, io, filename, file_id)

class YTDataContainerDataset(YTDataset):
    """Dataset for saved geometric data containers."""
    _index_class = ParticleIndex
    _file_class = YTDataHDF5File
    _field_info_class = YTDataContainerFieldInfo
    _suffix = ".h5"
    fluid_types = ("grid", "gas", "deposit", "index")

    def __init__(self, filename, dataset_type="ytdatacontainer_hdf5",
                 n_ref = 16, over_refine_factor = 1, units_override=None,
                 unit_system="cgs"):
        self.n_ref = n_ref
        self.over_refine_factor = over_refine_factor
        super(YTDataContainerDataset, self).__init__(filename, dataset_type,
            units_override=units_override, unit_system=unit_system)

    def _parse_parameter_file(self):
        super(YTDataContainerDataset, self)._parse_parameter_file()
        self.particle_types_raw = tuple(self.num_particles.keys())
        self.particle_types = self.particle_types_raw
        self.filename_template = self.parameter_filename
        self.file_count = 1
        nz = 1 << self.over_refine_factor
        self.domain_dimensions = np.ones(3, "int32") * nz

    def _setup_gas_alias(self):
        "Alias the grid type to gas by making a particle union."

        if "grid" in self.particle_types and \
          "gas" not in self.particle_types:
            pu = ParticleUnion("gas", ["grid"])
            self.add_particle_union(pu)
        # We have to alias this because particle unions only
        # cover the field_list.
        self.field_info.alias(("gas", "cell_volume"), ("grid", "cell_volume"))

    _data_obj = None
    @property
    def data(self):
        """
        Return a data container configured like the original used to
        create this dataset.
        """

        if self._data_obj is None:
            # Some data containers can't be reconstructed in the same way
            # since this is now particle-like data.
            data_type = self.parameters.get("data_type")
            container_type = self.parameters.get("container_type")
            ex_container_type = ["cutting", "proj", "ray", "slice", "cut_region"]
            if data_type == "yt_light_ray" or container_type in ex_container_type:
                mylog.info("Returning an all_data data container.")
                return self.all_data()

            my_obj = getattr(self, self.parameters["container_type"])
            my_args = [self.parameters[con_arg]
                       for con_arg in self.parameters["con_args"]]
            self._data_obj = my_obj(*my_args)
        return self._data_obj

    @classmethod
    def _is_valid(self, *args, **kwargs):
        if not args[0].endswith(".h5"): return False
        with h5py.File(args[0], "r") as f:
            data_type = parse_h5_attr(f, "data_type")
            cont_type = parse_h5_attr(f, "container_type")
            if data_type is None:
                return False
            if data_type == "yt_data_container" and \
                cont_type not in _grid_data_containers:
                return True
        return False

class YTDataLightRayDataset(YTDataContainerDataset):
    """Dataset for saved LightRay objects."""

    def _parse_parameter_file(self):
        super(YTDataLightRayDataset, self)._parse_parameter_file()
        self._restore_light_ray_solution()

    def _restore_light_ray_solution(self):
        """
        Restore all information associate with the light ray solution
        to its original form.
        """
        key = "light_ray_solution"
        self.light_ray_solution = []
        lrs_fields = [par for par in self.parameters \
                      if key in par and not par.endswith("_units")]
        if len(lrs_fields) == 0:
            return
        self.light_ray_solution = \
          [{} for val in self.parameters[lrs_fields[0]]]
        for sp3 in ["unique_identifier", "filename"]:
            ksp3 = "%s_%s" % (key, sp3)
            if ksp3 not in lrs_fields:
                continue
            self.parameters[ksp3] = self.parameters[ksp3].astype(str)
        for field in lrs_fields:
            field_name = field[len(key)+1:]
            for i in range(self.parameters[field].shape[0]):
                self.light_ray_solution[i][field_name] = \
                  self.parameters[field][i]

    @classmethod
    def _is_valid(self, *args, **kwargs):
        if not args[0].endswith(".h5"): return False
        with h5py.File(args[0], "r") as f:
            data_type = parse_h5_attr(f, "data_type")
            if data_type in ["yt_light_ray"]:
                return True
        return False

class YTSpatialPlotDataset(YTDataContainerDataset):
    """Dataset for saved slices and projections."""
    _field_info_class = YTGridFieldInfo

    def __init__(self, *args, **kwargs):
        super(YTSpatialPlotDataset, self).__init__(
            *args, dataset_type="ytspatialplot_hdf5", **kwargs)

    def _parse_parameter_file(self):
        super(YTSpatialPlotDataset, self)._parse_parameter_file()
        if self.parameters["container_type"] == "proj":
            if isinstance(self.parameters["weight_field"], string_types) and \
              self.parameters["weight_field"] == "None":
                self.parameters["weight_field"] = None
            elif isinstance(self.parameters["weight_field"], np.ndarray):
                self.parameters["weight_field"] = \
                  tuple(self.parameters["weight_field"])

    @classmethod
    def _is_valid(self, *args, **kwargs):
        if not args[0].endswith(".h5"): return False
        with h5py.File(args[0], "r") as f:
            data_type = parse_h5_attr(f, "data_type")
            cont_type = parse_h5_attr(f, "container_type")
            if data_type == "yt_data_container" and \
                cont_type in ["cutting", "proj", "slice"]:
                return True
        return False

class YTGrid(AMRGridPatch):
    _id_offset = 0
    def __init__(self, gid, index, filename=None):
        AMRGridPatch.__init__(self, gid, filename=filename, index=index)
        self._children_ids = []
        self._parent_id = -1
        self.Level = 0
        self.LeftEdge = self.index.ds.domain_left_edge
        self.RightEdge = self.index.ds.domain_right_edge

    def __getitem__(self, key):
        tr = super(AMRGridPatch, self).__getitem__(key)
        try:
            fields = self._determine_fields(key)
        except YTFieldTypeNotFound:
            return tr
        finfo = self.ds._get_field_info(*fields[0])
        if not finfo.particle_type:
            return tr.reshape(self.ActiveDimensions[:self.ds.dimensionality])
        return tr

    @property
    def Parent(self):
        return None

    @property
    def Children(self):
        return []

class YTDataHierarchy(GridIndex):
    def __init__(self, ds, dataset_type = None):
        self.dataset_type = dataset_type
        self.float_type = 'float64'
        self.dataset = weakref.proxy(ds)
        self.directory = os.getcwd()
        super(YTDataHierarchy, self).__init__(ds, dataset_type)

    def _count_grids(self):
        self.num_grids = 1

    def _parse_index(self):
        self.grid_dimensions[:] = self.ds.domain_dimensions
        self.grid_left_edge[:] = self.ds.domain_left_edge
        self.grid_right_edge[:] = self.ds.domain_right_edge
        self.grid_levels[:] = np.zeros(self.num_grids)
        self.grid_procs = np.zeros(self.num_grids)
        self.grid_particle_count[:] = sum(self.ds.num_particles.values())
        self.grids = []
        for gid in range(self.num_grids):
            self.grids.append(self.grid(gid, self))
            self.grids[gid].Level = self.grid_levels[gid, 0]
        self.max_level = self.grid_levels.max()
        temp_grids = np.empty(self.num_grids, dtype='object')
        for i, grid in enumerate(self.grids):
            grid.filename = self.ds.parameter_filename
            grid._prepare_grid()
            grid.proc_num = self.grid_procs[i]
            temp_grids[i] = grid
        self.grids = temp_grids

    def _detect_output_fields(self):
        self.field_list = []
        self.ds.field_units = self.ds.field_units or {}
        with h5py.File(self.ds.parameter_filename, "r") as f:
            for group in f:
                for field in f[group]:
                    field_name = (str(group), str(field))
                    self.field_list.append(field_name)
                    self.ds.field_units[field_name] = \
                      parse_h5_attr(f[group][field], "units")

class YTGridHierarchy(YTDataHierarchy):
    grid = YTGrid

    def _populate_grid_objects(self):
        for g in self.grids:
            g._setup_dx()
        self.max_level = self.grid_levels.max()

class YTGridDataset(YTDataset):
    """Dataset for saved covering grids, arbitrary grids, and FRBs."""
    _index_class = YTGridHierarchy
    _field_info_class = YTGridFieldInfo
    _dataset_type = 'ytgridhdf5'
    geometry = "cartesian"
    default_fluid_type = "grid"
    fluid_types = ("grid", "gas", "deposit", "index")

    def __init__(self, filename, unit_system="cgs"):
        super(YTGridDataset, self).__init__(filename, self._dataset_type,
                                            unit_system=unit_system)
        self.data = self.index.grids[0]

    def _parse_parameter_file(self):
        super(YTGridDataset, self)._parse_parameter_file()
        self.num_particles.pop(self.default_fluid_type, None)
        self.particle_types_raw = tuple(self.num_particles.keys())
        self.particle_types = self.particle_types_raw

        # correct domain dimensions for the covering grid dimension
        self.base_domain_left_edge = self.domain_left_edge
        self.base_domain_right_edge = self.domain_right_edge
        self.base_domain_dimensions = self.domain_dimensions

        if self.container_type in _grid_data_containers:
            self.domain_left_edge = self.parameters["left_edge"]

            if "level" in self.parameters["con_args"]:
                dx = (self.base_domain_right_edge -
                      self.base_domain_left_edge) / \
                  (self.domain_dimensions *
                   self.refine_by**self.parameters["level"])
                self.domain_right_edge = self.domain_left_edge + \
                  self.parameters["ActiveDimensions"] * dx
                self.domain_dimensions = \
                  ((self.domain_right_edge -
                    self.domain_left_edge) / dx).astype(int)
            else:
                self.domain_right_edge = self.parameters["right_edge"]
                self.domain_dimensions = self.parameters["ActiveDimensions"]
                dx = (self.domain_right_edge - self.domain_left_edge) / \
                  self.domain_dimensions

            self.periodicity = \
              np.abs(self.domain_left_edge -
                     self.base_domain_left_edge) < 0.5 * dx
            self.periodicity &= \
            np.abs(self.domain_right_edge -
                   self.base_domain_right_edge) < 0.5 * dx

        elif self.data_type == "yt_frb":
            dle = self.domain_left_edge
            self.domain_left_edge = uconcatenate(
                [self.parameters["left_edge"].to(dle.units), [0] * dle.uq])
            dre = self.domain_right_edge
            self.domain_right_edge = uconcatenate(
                [self.parameters["right_edge"].to(dre.units), [1] * dre.uq])
            self.domain_dimensions = \
              np.concatenate([self.parameters["ActiveDimensions"], [1]])

    def _setup_gas_alias(self):
        "Alias the grid type to gas with a field alias."

        for ftype, field in self.field_list:
            if ftype == "grid":
                self.field_info.alias(("gas", field), ("grid", field))

    @classmethod
    def _is_valid(self, *args, **kwargs):
        if not args[0].endswith(".h5"): return False
        with h5py.File(args[0], "r") as f:
            data_type = parse_h5_attr(f, "data_type")
            cont_type = parse_h5_attr(f, "container_type")
            if data_type == "yt_frb":
                return True
            if data_type == "yt_data_container" and \
                cont_type in _grid_data_containers:
                return True
        return False

class YTNonspatialGrid(AMRGridPatch):
    _id_offset = 0
    def __init__(self, gid, index, filename=None):
        super(YTNonspatialGrid, self).__init__(gid, filename=filename, index=index)
        self._children_ids = []
        self._parent_id = -1
        self.Level = 0
        self.LeftEdge = self.index.ds.domain_left_edge
        self.RightEdge = self.index.ds.domain_right_edge

    def __repr__(self):
        return "YTNonspatialGrid"

    def __getitem__(self, key):
        tr = super(AMRGridPatch, self).__getitem__(key)
        try:
            fields = self._determine_fields(key)
        except YTFieldTypeNotFound:
            return tr
        self.ds._get_field_info(*fields[0])
        return tr

    def get_data(self, fields=None):
        if fields is None: return
        nfields = []
        apply_fields = defaultdict(list)
        for field in self._determine_fields(fields):
            if field[0] in self.ds.filtered_particle_types:
                f = self.ds.known_filters[field[0]]
                apply_fields[field[0]].append(
                    (f.filtered_type, field[1]))
            else:
                nfields.append(field)
        for filter_type in apply_fields:
            f = self.ds.known_filters[filter_type]
            with f.apply(self):
                self.get_data(apply_fields[filter_type])
        fields = nfields
        if len(fields) == 0: return
        # Now we collect all our fields
        # Here is where we need to perform a validation step, so that if we
        # have a field requested that we actually *can't* yet get, we put it
        # off until the end.  This prevents double-reading fields that will
        # need to be used in spatial fields later on.
        fields_to_get = []
        # This will be pre-populated with spatial fields
        fields_to_generate = []
        for field in self._determine_fields(fields):
            if field in self.field_data: continue
            finfo = self.ds._get_field_info(*field)
            try:
                finfo.check_available(self)
            except NeedsGridType:
                fields_to_generate.append(field)
                continue
            fields_to_get.append(field)
        if len(fields_to_get) == 0 and len(fields_to_generate) == 0:
            return
        elif self._locked is True:
            raise GenerationInProgress(fields)
        # Track which ones we want in the end
        ofields = set(list(self.field_data.keys())
                    + fields_to_get
                    + fields_to_generate)
        # At this point, we want to figure out *all* our dependencies.
        fields_to_get = self._identify_dependencies(fields_to_get,
            self._spatial)
        # We now split up into readers for the types of fields
        fluids, particles = [], []
        finfos = {}
        for ftype, fname in fields_to_get:
            finfo = self.ds._get_field_info(ftype, fname)
            finfos[ftype, fname] = finfo
            if finfo.particle_type:
                particles.append((ftype, fname))
            elif (ftype, fname) not in fluids:
                fluids.append((ftype, fname))

        # The _read method will figure out which fields it needs to get from
        # disk, and return a dict of those fields along with the fields that
        # need to be generated.
        read_fluids, gen_fluids = self.index._read_fluid_fields(
                                        fluids, self, self._current_chunk)
        for f, v in read_fluids.items():
            convert = True
            if v.dtype != np.float64:
                if finfos[f].units == "":
                    self.field_data[f] = v
                    convert = False
                else:
                    v = v.astype(np.float64)
            if convert:
                self.field_data[f] = self.ds.arr(v, input_units = finfos[f].units)
                self.field_data[f].convert_to_units(finfos[f].output_units)

        read_particles, gen_particles = self.index._read_fluid_fields(
                                        particles, self, self._current_chunk)
        for f, v in read_particles.items():
            convert = True
            if v.dtype != np.float64:
                if finfos[f].units == "":
                    self.field_data[f] = v
                    convert = False
                else:
                    v = v.astype(np.float64)
            if convert:
                self.field_data[f] = self.ds.arr(v, input_units = finfos[f].units)
                self.field_data[f].convert_to_units(finfos[f].output_units)

        fields_to_generate += gen_fluids + gen_particles
        self._generate_fields(fields_to_generate)
        for field in list(self.field_data.keys()):
            if field not in ofields:
                self.field_data.pop(field)

    @property
    def Parent(self):
        return None

    @property
    def Children(self):
        return []

class YTNonspatialHierarchy(YTDataHierarchy):
    grid = YTNonspatialGrid

    def _populate_grid_objects(self):
        for g in self.grids:
            g._setup_dx()
            # this is non-spatial, so remove the code_length units
            g.dds = self.ds.arr(g.dds.d, "")
            g.ActiveDimensions = self.ds.domain_dimensions
        self.max_level = self.grid_levels.max()

    def _read_fluid_fields(self, fields, dobj, chunk = None):
        if len(fields) == 0: return {}, []
        fields_to_read, fields_to_generate = self._split_fields(fields)
        if len(fields_to_read) == 0:
            return {}, fields_to_generate
        selector = dobj.selector
        fields_to_return = self.io._read_fluid_selection(
            dobj,
            selector,
            fields_to_read)
        return fields_to_return, fields_to_generate

class YTNonspatialDataset(YTGridDataset):
    """Dataset for general array data."""
    _index_class = YTNonspatialHierarchy
    _field_info_class = YTGridFieldInfo
    _dataset_type = 'ytnonspatialhdf5'
    geometry = "cartesian"
    default_fluid_type = "data"
    fluid_types = ("data", "gas")

    def _parse_parameter_file(self):
        super(YTGridDataset, self)._parse_parameter_file()
        self.num_particles.pop(self.default_fluid_type, None)
        self.particle_types_raw = tuple(self.num_particles.keys())
        self.particle_types = self.particle_types_raw

    def _set_derived_attrs(self):
        # set some defaults just to make things go
        default_attrs = {
            "dimensionality": 3,
            "domain_dimensions": np.ones(3, dtype="int"),
            "domain_left_edge": np.zeros(3),
            "domain_right_edge": np.ones(3),
            "periodicity": np.ones(3, dtype="bool")
        }
        for att, val in default_attrs.items():
            if getattr(self, att, None) is None:
                setattr(self, att, val)

    def _setup_classes(self):
        # We don't allow geometric selection for non-spatial datasets
        self.objects = []

    @parallel_root_only
    def print_key_parameters(self):
        for a in ["current_time", "domain_dimensions", "domain_left_edge",
                  "domain_right_edge", "cosmological_simulation"]:
            v = getattr(self, a)
            if v is not None: mylog.info("Parameters: %-25s = %s", a, v)
        if hasattr(self, "cosmological_simulation") and \
           getattr(self, "cosmological_simulation"):
            for a in ["current_redshift", "omega_lambda", "omega_matter",
                      "hubble_constant"]:
                v = getattr(self, a)
                if v is not None: mylog.info("Parameters: %-25s = %s", a, v)
        mylog.warn("Geometric data selection not available for this dataset type.")

    @classmethod
    def _is_valid(self, *args, **kwargs):
        if not args[0].endswith(".h5"): return False
        with h5py.File(args[0], "r") as f:
            data_type = parse_h5_attr(f, "data_type")
            if data_type == "yt_array_data":
                return True
        return False

class YTProfileDataset(YTNonspatialDataset):
    """Dataset for saved profile objects."""
    def __init__(self, filename, unit_system="cgs"):
        super(YTProfileDataset, self).__init__(filename,
                                               unit_system=unit_system)

    @property
    def profile(self):
        if self.dimensionality == 1:
            return Profile1DFromDataset(self)
        if self.dimensionality == 2:
            return Profile2DFromDataset(self)
        if self.dimensionality == 3:
            return Profile3DFromDataset(self)
        return None

    def _parse_parameter_file(self):
        super(YTGridDataset, self)._parse_parameter_file()

        if isinstance(self.parameters["weight_field"], string_types) and \
          self.parameters["weight_field"] == "None":
            self.parameters["weight_field"] = None
        elif isinstance(self.parameters["weight_field"], np.ndarray):
            self.parameters["weight_field"] = \
              tuple(self.parameters["weight_field"].astype(str))

        for a in ["profile_dimensions"] + \
          ["%s_%s" % (ax, attr)
           for ax in "xyz"[:self.dimensionality]
           for attr in ["log"]]:
            setattr(self, a, self.parameters[a])

        self.base_domain_left_edge = self.domain_left_edge
        self.base_domain_right_edge = self.domain_right_edge
        self.base_domain_dimensions = self.domain_dimensions

        domain_dimensions = np.ones(3, dtype="int")
        domain_dimensions[:self.dimensionality] = \
          self.profile_dimensions
        self.domain_dimensions = domain_dimensions
        domain_left_edge = np.zeros(3)
        domain_right_edge = np.ones(3)
        for i, ax in enumerate("xyz"[:self.dimensionality]):
            range_name = "%s_range" % ax
            my_range = self.parameters[range_name]
            if getattr(self, "%s_log" % ax, False):
                my_range = np.log10(my_range)
            domain_left_edge[i] = my_range[0]
            domain_right_edge[i] = my_range[1]
            setattr(self, range_name, self.parameters[range_name])

            bin_field = "%s_field" % ax
            if isinstance(self.parameters[bin_field], string_types) and \
              self.parameters[bin_field] == "None":
                self.parameters[bin_field] = None
            elif isinstance(self.parameters[bin_field], np.ndarray):
                self.parameters[bin_field] = \
                  tuple(self.parameters[bin_field].astype(str))
            setattr(self, bin_field, self.parameters[bin_field])
        self.domain_left_edge = domain_left_edge
        self.domain_right_edge = domain_right_edge

    def _setup_gas_alias(self):
        "Alias the grid type to gas with a field alias."
        for ftype, field in self.field_list:
            if ftype == "data":
                self.field_info.alias(("gas", field), (ftype, field))

    def create_field_info(self):
        super(YTProfileDataset, self).create_field_info()
        if self.parameters["weight_field"] is not None:
            self.field_info.alias(self.parameters["weight_field"],
                                  (self.default_fluid_type, "weight"))

    def _set_derived_attrs(self):
        self.domain_center = 0.5 * (self.domain_right_edge +
                                    self.domain_left_edge)
        self.domain_width = self.domain_right_edge - \
          self.domain_left_edge

    def print_key_parameters(self):
        if is_root():
            mylog.info("YTProfileDataset")
            for a in ["dimensionality", "profile_dimensions"] + \
              ["%s_%s" % (ax, attr)
               for ax in "xyz"[:self.dimensionality]
               for attr in ["field", "range", "log"]]:
                v = getattr(self, a)
                mylog.info("Parameters: %-25s = %s", a, v)
        super(YTProfileDataset, self).print_key_parameters()

    @classmethod
    def _is_valid(self, *args, **kwargs):
        if not args[0].endswith(".h5"): return False
        with h5py.File(args[0], "r") as f:
            data_type = parse_h5_attr(f, "data_type")
            if data_type == "yt_profile":
                return True
        return False

class YTClumpContainer(TreeContainer):
    def __init__(self, clump_id, global_id, parent_id,
                 contour_key, contour_id, ds=None):
        self.clump_id = clump_id
        self.global_id = global_id
        self.parent_id = parent_id
        self.contour_key = contour_key
        self.contour_id = contour_id
        self.parent = None
        self.ds = ds
        TreeContainer.__init__(self)

    def add_child(self, child):
        if self.children is None:
            self.children = []
        self.children.append(child)
        child.parent = self

    def __repr__(self):
        return "Clump[%d]" % self.clump_id

    def __getitem__(self, field):
        g = self.ds.data
        f = g._determine_fields(field)[0]
        if f[0] == "clump":
            return g[f][self.global_id]
        if self.contour_id == -1:
            return g[f]
        cfield = (f[0], "contours_%s" % self.contour_key.decode('utf-8'))
        if f[0] == "grid":
            return g[f][g[cfield] == self.contour_id]
        return self.parent[f][g[cfield] == self.contour_id]

class YTClumpTreeDataset(YTNonspatialDataset):
    """Dataset for saved clump-finder data."""
    def __init__(self, filename, unit_system="cgs"):
        super(YTClumpTreeDataset, self).__init__(filename,
                                                 unit_system=unit_system)
        self._load_tree()

    def _load_tree(self):
        my_tree = {}
        for i, clump_id in enumerate(self.data[("clump", "clump_id")]):
            my_tree[clump_id] = YTClumpContainer(
                clump_id, i, self.data["clump", "parent_id"][i],
                self.data["clump", "contour_key"][i],
                self.data["clump", "contour_id"][i], self)
        for clump in my_tree.values():
            if clump.parent_id == -1:
                self.tree = clump
            else:
                parent = my_tree[clump.parent_id]
                parent.add_child(clump)

    _leaves = None
    @property
    def leaves(self):
        if self._leaves is None:
            self._leaves = []
            for clump in self.tree:
                if clump.children is None:
                    self._leaves.append(clump)
        return self._leaves

    @classmethod
    def _is_valid(self, *args, **kwargs):
        if not args[0].endswith(".h5"): return False
        with h5py.File(args[0], "r") as f:
            data_type = parse_h5_attr(f, "data_type")
            if data_type is None:
                return False
            if data_type == "yt_clump_tree":
                return True
        return False



"""
Fields specific to Streaming data



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.fields.field_info_container import \
    FieldInfoContainer


class StreamFieldInfo(FieldInfoContainer):
    known_other_fields = (
        ("density", ("code_mass/code_length**3", ["density"], None)),
        ("dark_matter_density", ("code_mass/code_length**3", ["dark_matter_density"], None)),
        ("number_density", ("1/code_length**3", ["number_density"], None)),
        ("pressure", ("dyne/code_length**2", ["pressure"], None)),
        ("thermal_energy", ("erg / g", ["thermal_energy"], None)),
        ("temperature", ("K", ["temperature"], None)),
        ("velocity_x", ("code_length/code_time", ["velocity_x"], None)),
        ("velocity_y", ("code_length/code_time", ["velocity_y"], None)),
        ("velocity_z", ("code_length/code_time", ["velocity_z"], None)),
        ("magnetic_field_x", ("gauss", [], None)),
        ("magnetic_field_y", ("gauss", [], None)),
        ("magnetic_field_z", ("gauss", [], None)),
        ("radiation_acceleration_x", ("code_length/code_time**2", ["radiation_acceleration_x"], None)),
        ("radiation_acceleration_y", ("code_length/code_time**2", ["radiation_acceleration_y"], None)),
        ("radiation_acceleration_z", ("code_length/code_time**2", ["radiation_acceleration_z"], None)),
        ("metallicity", ("Zsun", ["metallicity"], None)),

        # We need to have a bunch of species fields here, too
        ("metal_density",   ("code_mass/code_length**3", ["metal_density"], None)),
        ("hi_density",      ("code_mass/code_length**3", ["hi_density"], None)),
        ("hii_density",     ("code_mass/code_length**3", ["hii_density"], None)),
        ("h2i_density",     ("code_mass/code_length**3", ["h2i_density"], None)),
        ("h2ii_density",    ("code_mass/code_length**3", ["h2ii_density"], None)),
        ("h2m_density",     ("code_mass/code_length**3", ["h2m_density"], None)),
        ("hei_density",     ("code_mass/code_length**3", ["hei_density"], None)),
        ("heii_density",    ("code_mass/code_length**3", ["heii_density"], None)),
        ("heiii_density",   ("code_mass/code_length**3", ["heiii_density"], None)),
        ("hdi_density",     ("code_mass/code_length**3", ["hdi_density"], None)),
        ("di_density",      ("code_mass/code_length**3", ["di_density"], None)),
        ("dii_density",     ("code_mass/code_length**3", ["dii_density"], None)),
    )

    known_particle_fields = (
        ("particle_position", ("code_length", ["particle_position"], None)),
        ("particle_position_x", ("code_length", ["particle_position_x"], None)),
        ("particle_position_y", ("code_length", ["particle_position_y"], None)),
        ("particle_position_z", ("code_length", ["particle_position_z"], None)),
        ("particle_velocity", ("code_length/code_time", ["particle_velocity"], None)),
        ("particle_velocity_x", ("code_length/code_time", ["particle_velocity_x"], None)),
        ("particle_velocity_y", ("code_length/code_time", ["particle_velocity_y"], None)),
        ("particle_velocity_z", ("code_length/code_time", ["particle_velocity_z"], None)),
        ("particle_index", ("", ["particle_index"], None)),
        ("particle_gas_density", ("code_mass/code_length**3", ["particle_gas_density"], None)),
        ("particle_gas_temperature", ("K", ["particle_gas_temperature"], None)),
        ("particle_mass", ("code_mass", ["particle_mass"], None)),
        ("smoothing_length", ("code_length", ["smoothing_length"], None)),
        ("density", ("code_mass/code_length**3", ["density"], None)),
        ("temperature", ("code_temperature", ["temperature"], None)),
        ("creation_time", ("code_time", ["creation_time"], None)),
    )

    def setup_fluid_fields(self):
        from yt.fields.magnetic_field import \
            setup_magnetic_field_aliases
        for field in self.ds.stream_handler.field_units:
            if field[0] in self.ds.particle_types:
                continue
            units = self.ds.stream_handler.field_units[field]
            if units != '': 
                self.add_output_field(field, sampling_type="cell", units=units)
        setup_magnetic_field_aliases(self, "stream", ["magnetic_field_%s" % ax for ax in "xyz"])

    def add_output_field(self, name, sampling_type, **kwargs):
        if name in self.ds.stream_handler.field_units:
            kwargs['units'] = self.ds.stream_handler.field_units[name]
        super(StreamFieldInfo, self).add_output_field(name, sampling_type, **kwargs)



"""
Athena frontend tests



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.testing import \
    assert_equal, \
    requires_file, \
    assert_allclose_units, \
    disable_dataset_cache
from yt.utilities.answer_testing.framework import \
    requires_ds, \
    small_patch_amr, \
    data_dir_load
from yt.frontends.athena.api import AthenaDataset
from yt.convenience import load

import yt.units as u

_fields_cloud = ("scalar[0]", "density", "total_energy")

cloud = "ShockCloud/id0/Cloud.0050.vtk"
@requires_ds(cloud)
def test_cloud():
    ds = data_dir_load(cloud)
    assert_equal(str(ds), "Cloud.0050")
    for test in small_patch_amr(ds, _fields_cloud):
        test_cloud.__name__ = test.description
        yield test

_fields_blast = ("temperature", "density", "velocity_magnitude")

blast = "MHDBlast/id0/Blast.0100.vtk"
@requires_ds(blast)
def test_blast():
    ds = data_dir_load(blast)
    assert_equal(str(ds), "Blast.0100")
    for test in small_patch_amr(ds, _fields_blast):
        test_blast.__name__ = test.description
        yield test

uo_blast = {
    'length_unit': (1.0, 'pc'),
    'mass_unit': (2.38858753789e-24, 'g/cm**3*pc**3'),
    'time_unit': (1.0, 's*pc/km'),
}

@requires_file(blast)
def test_blast_override():
    # verify that overriding units causes derived unit values to be updated.
    # see issue #1259
    ds = load(blast, units_override=uo_blast)
    assert_equal(float(ds.magnetic_unit.in_units('gauss')),
                 5.478674679698131e-07)

uo_stripping = {"time_unit":3.086e14,
                "length_unit":8.0236e22,
                "mass_unit":9.999e-30*8.0236e22**3}

_fields_stripping = ("temperature", "density", "specific_scalar[0]")

stripping = "RamPressureStripping/id0/rps.0062.vtk"
@requires_ds(stripping, big_data=True)
def test_stripping():
    ds = data_dir_load(stripping, kwargs={"units_override":uo_stripping})
    assert_equal(str(ds), "rps.0062")
    for test in small_patch_amr(ds, _fields_stripping):
        test_stripping.__name__ = test.description
        yield test

sloshing = "MHDSloshing/virgo_low_res.0054.vtk"

uo_sloshing = {"length_unit": (1.0,"Mpc"),
               "time_unit": (1.0,"Myr"),
               "mass_unit": (1.0e14,"Msun")}

@requires_file(sloshing)
@disable_dataset_cache
def test_nprocs():
    ds1 = load(sloshing, units_override=uo_sloshing)
    sp1 = ds1.sphere("c", (100.,"kpc"))
    prj1 = ds1.proj("density",0)
    ds2 = load(sloshing, units_override=uo_sloshing, nprocs=8)
    sp2 = ds2.sphere("c", (100.,"kpc"))
    prj2 = ds1.proj("density",0)

    ds3 = load(sloshing, parameters=uo_sloshing)
    assert_equal(ds3.length_unit, u.Mpc)
    assert_equal(ds3.time_unit, u.Myr)
    assert_equal(ds3.mass_unit, 1e14*u.Msun)

    assert_equal(sp1.quantities.extrema("pressure"),
                 sp2.quantities.extrema("pressure"))
    assert_allclose_units(sp1.quantities.total_quantity("pressure"),
                          sp2.quantities.total_quantity("pressure"))
    for ax in "xyz":
        assert_equal(sp1.quantities.extrema("velocity_%s" % ax),
                     sp2.quantities.extrema("velocity_%s" % ax))
    assert_allclose_units(sp1.quantities.bulk_velocity(),
                          sp2.quantities.bulk_velocity())
    assert_equal(prj1["density"], prj2["density"])

@requires_file(cloud)
def test_AthenaDataset():
    assert isinstance(data_dir_load(cloud), AthenaDataset)



from yt.testing import fake_random_ds
from yt.data_objects.profiles import create_profile
from numpy.random import uniform

def test_update_data():
    ds = fake_random_ds(64, nprocs=8)
    ds.index
    dims = (32,32,32)
    grid_data = [{"temperature": uniform(size=dims)}
                 for i in range(ds.index.num_grids)]
    ds.index.update_data(grid_data)
    prj = ds.proj("temperature", 2)
    prj["temperature"]
    dd = ds.all_data()
    profile = create_profile(dd, "density", "temperature", 10)
    profile["temperature"]



import numpy as np

from yt.testing import assert_equal
from yt.utilities.lib.misc_utilities import fill_region

NDIM = 32

def test_fill_region():
    for level in range(2):
        rf = 2**level
        output_fields = [np.zeros((NDIM*rf,NDIM*rf,NDIM*rf), "float64")
                         for i in range(3)]
        input_fields = [np.empty(NDIM**3, "float64")
                         for i in range(3)]
        v = np.mgrid[0.0:1.0:NDIM*1j, 0.0:1.0:NDIM*1j, 0.0:1.0:NDIM*1j]
        input_fields[0][:] = v[0].ravel()
        input_fields[1][:] = v[1].ravel()
        input_fields[2][:] = v[2].ravel()
        left_index = np.zeros(3, "int64")
        ipos = np.empty((NDIM**3, 3), dtype="int64")
        ind = np.indices((NDIM,NDIM,NDIM))
        ipos[:,0] = ind[0].ravel()
        ipos[:,1] = ind[1].ravel()
        ipos[:,2] = ind[2].ravel()
        ires = np.zeros(NDIM*NDIM*NDIM, "int64")
        ddims = np.array([NDIM, NDIM, NDIM], dtype="int64") * rf
        fill_region(input_fields, output_fields, level,
                    left_index, ipos, ires, ddims,
                    np.array([2, 2, 2], dtype="i8"))
        for r in range(level + 1):
            for o, i in zip(output_fields, v):
                assert_equal( o[r::rf,r::rf,r::rf], i)




"""
This is where we define a handful of vector operations for fields.

"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from .derived_field import \
    ValidateParameter, \
    ValidateSpatial

from yt.utilities.math_utils import \
    get_sph_r_component, \
    get_sph_theta_component, \
    get_sph_phi_component, \
    get_cyl_r_component, \
    get_cyl_z_component, \
    get_cyl_theta_component

from yt.funcs import just_one

from yt.utilities.lib.misc_utilities import obtain_rv_vec

def create_magnitude_field(registry, basename, field_units,
                           ftype="gas", slice_info=None,
                           validators=None, particle_type=False):

    field_components = [(ftype, "%s_%s" % (basename, ax)) for ax in 'xyz']

    def _magnitude(field, data):
        fn = field_components[0]
        mag = data[fn] * data[fn]
        for idim in range(1, registry.ds.dimensionality):
            fn = field_components[idim]
            mag += data[fn] * data[fn]
        return np.sqrt(mag)

    registry.add_field((ftype, "%s_magnitude" % basename), sampling_type="cell", 
                       function=_magnitude, units=field_units,
                       validators=validators, particle_type=particle_type)

def create_squared_field(registry, basename, field_units,
                         ftype="gas", slice_info=None,
                         validators=None, particle_type=False):

    field_components = [(ftype, "%s_%s" % (basename, ax)) for ax in 'xyz']

    def _squared(field, data):
        fn = field_components[0]
        squared = data[fn] * data[fn]
        for idim in range(1, registry.ds.dimensionality):
            fn = field_components[idim]
            squared += data[fn] * data[fn]
        return squared

    registry.add_field((ftype, "%s_squared" % basename), sampling_type="cell", 
                       function=_squared, units=field_units,
                       validators=validators, particle_type=particle_type)

def create_vector_fields(registry, basename, field_units,
                         ftype="gas", slice_info=None):
    from yt.units.unit_object import Unit
    # slice_info would be the left, the right, and the factor.
    # For example, with the old Enzo-ZEUS fields, this would be:
    # slice(None, -2, None)
    # slice(1, -1, None)
    # 1.0
    # Otherwise, we default to a centered difference.
    if slice_info is None:
        sl_left = slice(None, -2, None)
        sl_right = slice(2, None, None)
        div_fac = 2.0
    else:
        sl_left, sl_right, div_fac = slice_info

    xn, yn, zn = [(ftype, "%s_%s" % (basename, ax)) for ax in 'xyz']

    # Is this safe?
    if registry.ds.dimensionality < 3:
        zn = ("index", "zeros")
    if registry.ds.dimensionality < 2:
        yn = ("index", "zeros")

    create_magnitude_field(registry, basename, field_units,
                           ftype=ftype, slice_info=slice_info)

    def _spherical_radius_component(field, data):
        """The spherical radius component of the vector field

        Relative to the coordinate system defined by the *normal* vector,
        *center*, and *bulk_* field parameters.
        """
        normal = data.get_field_parameter("normal")
        vectors = obtain_rv_vec(data, (xn, yn, zn),
                                "bulk_%s" % basename)
        theta = data['index', 'spherical_theta']
        phi = data['index', 'spherical_phi']
        rv = get_sph_r_component(vectors, theta, phi, normal)
        # Now, anywhere that radius is in fact zero, we want to zero out our
        # return values.
        rv[np.isnan(theta)] = 0.0
        return rv

    registry.add_field((ftype, "%s_spherical_radius" % basename), sampling_type="cell", 
                       function=_spherical_radius_component,
                       units=field_units,
                       validators=[ValidateParameter("normal"),
                                   ValidateParameter("center"),
                                   ValidateParameter("bulk_%s" % basename)])

    def _radial(field, data):
        return data[ftype, "%s_spherical_radius" % basename]

    def _radial_absolute(field, data):
        return np.abs(data[ftype, "%s_spherical_radius" % basename])

    def _tangential(field, data):
        return np.sqrt(data[ftype, "%s_spherical_theta" % basename]**2.0 +
                       data[ftype, "%s_spherical_phi" % basename]**2.0)

    registry.add_field((ftype, "radial_%s" % basename), sampling_type="cell", 
                       function=_radial,
                       units=field_units,
                       validators=[ValidateParameter("normal"),
                                   ValidateParameter("center")])

    registry.add_field((ftype, "radial_%s_absolute" % basename), sampling_type="cell", 
                       function=_radial_absolute,
                       units=field_units)

    registry.add_field((ftype, "tangential_%s" % basename), sampling_type="cell", 
                       function=_tangential,
                       units=field_units)

    def _spherical_theta_component(field, data):
        """The spherical theta component of the vector field

        Relative to the coordinate system defined by the *normal* vector,
        *center*, and *bulk_* field parameters.
        """
        normal = data.get_field_parameter("normal")
        vectors = obtain_rv_vec(data, (xn, yn, zn),
                                "bulk_%s" % basename)
        theta = data["index", "spherical_theta"]
        phi = data["index", "spherical_phi"]
        return get_sph_theta_component(vectors, theta, phi, normal)

    registry.add_field((ftype, "%s_spherical_theta" % basename), sampling_type="cell", 
                       function=_spherical_theta_component,
                       units=field_units,
                       validators=[ValidateParameter("normal"),
                                   ValidateParameter("center"),
                                   ValidateParameter("bulk_%s" % basename)])

    def _spherical_phi_component(field, data):
        """The spherical phi component of the vector field

        Relative to the coordinate system defined by the *normal* vector,
        *center*, and *bulk_* field parameters.
        """
        normal = data.get_field_parameter("normal")
        vectors = obtain_rv_vec(data, (xn, yn, zn),
                                "bulk_%s" % basename)
        phi = data["index", "spherical_phi"]
        return get_sph_phi_component(vectors, phi, normal)

    registry.add_field((ftype, "%s_spherical_phi" % basename), sampling_type="cell", 
                       function=_spherical_phi_component,
                       units=field_units,
                       validators=[ValidateParameter("normal"),
                                   ValidateParameter("center"),
                                   ValidateParameter("bulk_%s" % basename)])

    def _cp_vectors(ax):
        def _cp_val(field, data):
            vec = data.get_field_parameter("cp_%s_vec" % (ax))
            bv = data.get_field_parameter("bulk_%s" % basename, None)
            if bv is None:
                bv = data.ds.arr(np.zeros(3), data[xn].units)
            tr  = (data[xn] - bv[0]) * vec.d[0]
            tr += (data[yn] - bv[1]) * vec.d[1]
            tr += (data[zn] - bv[2]) * vec.d[2]
            return tr
        return _cp_val

    registry.add_field((ftype, "cutting_plane_%s_x" % basename), sampling_type="cell", 
                       function=_cp_vectors('x'),
                       units=field_units)
    registry.add_field((ftype, "cutting_plane_%s_y" % basename), sampling_type="cell", 
                       function=_cp_vectors('y'),
                       units=field_units)
    registry.add_field((ftype, "cutting_plane_%s_z" % basename), sampling_type="cell", 
                       function=_cp_vectors('z'),
                       units=field_units)

    def _divergence(field, data):
        ds = div_fac * just_one(data["index", "dx"])
        f  = data[xn][sl_right,1:-1,1:-1]/ds
        f -= data[xn][sl_left ,1:-1,1:-1]/ds
        ds = div_fac * just_one(data["index", "dy"])
        f += data[yn][1:-1,sl_right,1:-1]/ds
        f -= data[yn][1:-1,sl_left ,1:-1]/ds
        ds = div_fac * just_one(data["index", "dz"])
        f += data[zn][1:-1,1:-1,sl_right]/ds
        f -= data[zn][1:-1,1:-1,sl_left ]/ds
        new_field = data.ds.arr(np.zeros(data[xn].shape, dtype=np.float64),
                                f.units)        
        new_field[1:-1,1:-1,1:-1] = f
        return new_field

    def _divergence_abs(field, data):
        return np.abs(data[ftype, "%s_divergence" % basename])

    field_units = Unit(field_units, registry=registry.ds.unit_registry)
    div_units = field_units / registry.ds.unit_system["length"]

    registry.add_field((ftype, "%s_divergence" % basename), sampling_type="cell", 
                       function=_divergence,
                       units=div_units,
                       validators=[ValidateSpatial(1)])
    
    registry.add_field((ftype, "%s_divergence_absolute" % basename), sampling_type="cell", 
                       function=_divergence_abs,
                       units=div_units)

    def _tangential_over_magnitude(field, data):
        tr = data[ftype, "tangential_%s" % basename] / \
             np.sqrt(data[ftype, "%s_spherical_radius" % basename]**2.0 +
                     data[ftype, "%s_spherical_theta" % basename]**2.0 +
                     data[ftype, "%s_spherical_phi" % basename]**2.0)
        return np.abs(tr)
    registry.add_field((ftype, "tangential_over_%s_magnitude" % basename), sampling_type="cell", 
             function=_tangential_over_magnitude,
             take_log=False)

    def _cylindrical_radius_component(field, data):
        """The cylindrical radius component of the vector field

        Relative to the coordinate system defined by the *normal* vector,
        *center*, and *bulk_* field parameters.
        """
        normal = data.get_field_parameter("normal")
        vectors = obtain_rv_vec(data, (xn, yn, zn),
                                "bulk_%s" % basename)
        theta = data["index", 'cylindrical_theta']
        return get_cyl_r_component(vectors, theta, normal)

    registry.add_field((ftype, "%s_cylindrical_radius" % basename), sampling_type="cell", 
                       function=_cylindrical_radius_component,
                       units=field_units,
                       validators=[ValidateParameter("normal")])

    def _cylindrical_radial(field, data):
        """This field is deprecated and will be removed in a future version"""
        return data[ftype, '%s_cylindrical_radius' % basename]

    registry.add_field((ftype, "cylindrical_radial_%s" % basename), sampling_type="cell", 
                       function=_cylindrical_radial,
                       units=field_units)

    def _cylindrical_radial_absolute(field, data):
        """This field is deprecated and will be removed in a future version"""
        return np.abs(data[ftype, '%s_cylindrical_radius' % basename])

    registry.add_field((ftype, "cylindrical_radial_%s_absolute" % basename), sampling_type="cell", 
                       function=_cylindrical_radial_absolute,
                       units=field_units,
                       validators=[ValidateParameter("normal")])

    def _cylindrical_theta_component(field, data):
        """The cylindrical theta component of the vector field

        Relative to the coordinate system defined by the *normal* vector,
        *center*, and *bulk_* field parameters.
        """
        normal = data.get_field_parameter("normal")
        vectors = obtain_rv_vec(data, (xn, yn, zn),
                                "bulk_%s" % basename)
        theta = data["index", 'cylindrical_theta'].copy()
        theta = np.tile(theta, (3,) + (1,)*len(theta.shape))
        return get_cyl_theta_component(vectors, theta, normal)

    registry.add_field((ftype, "%s_cylindrical_theta" % basename), sampling_type="cell", 
                       function=_cylindrical_theta_component,
                       units=field_units,
                       validators=[ValidateParameter("normal"),
                                   ValidateParameter("center"),
                                   ValidateParameter("bulk_%s" % basename)])

    def _cylindrical_tangential(field, data):
        """This field is deprecated and will be removed in a future release"""
        return data["%s_cylindrical_theta" % basename]

    def _cylindrical_tangential_absolute(field, data):
        """This field is deprecated and will be removed in a future release"""
        return np.abs(data['cylindrical_tangential_%s' % basename])

    registry.add_field((ftype, "cylindrical_tangential_%s" % basename), sampling_type="cell", 
                       function=_cylindrical_tangential,
                       units=field_units)

    registry.add_field((ftype, "cylindrical_tangential_%s_absolute" % basename), sampling_type="cell", 
                       function=_cylindrical_tangential_absolute,
                       units=field_units)

    def _cylindrical_z_component(field, data):
        """The cylindrical z component of the vector field

        Relative to the coordinate system defined by the *normal* vector,
        *center*, and *bulk_* field parameters.
        """
        normal = data.get_field_parameter("normal")
        vectors = obtain_rv_vec(data, (xn, yn, zn), "bulk_%s" % basename)
        return get_cyl_z_component(vectors, normal)

    registry.add_field((ftype, "%s_cylindrical_z" % basename), sampling_type="cell", 
                       function=_cylindrical_z_component,
                       units=field_units,
                       validators=[ValidateParameter("normal"),
                                   ValidateParameter("center"),
                                   ValidateParameter("bulk_%s" % basename)])


def create_averaged_field(registry, basename, field_units, ftype="gas",
                          slice_info=None, validators=None, weight="cell_mass"):

    if validators is None:
        validators = []
    validators += [ValidateSpatial(1, [(ftype, basename)])]

    def _averaged_field(field, data):
        nx, ny, nz = data[(ftype, basename)].shape
        new_field = data.ds.arr(np.zeros((nx-2, ny-2, nz-2), dtype=np.float64),
                                (just_one(data[(ftype, basename)]) *
                                 just_one(data[(ftype, weight)])).units)
        weight_field = data.ds.arr(np.zeros((nx-2, ny-2, nz-2),
                                            dtype=np.float64),
                                   data[(ftype, weight)].units)
        i_i, j_i, k_i = np.mgrid[0:3, 0:3, 0:3]

        for i, j, k in zip(i_i.ravel(), j_i.ravel(), k_i.ravel()):
            sl = [slice(i, nx-(2-i)), slice(j, ny-(2-j)), slice(k, nz-(2-k))]
            new_field += data[(ftype, basename)][sl] * data[(ftype, weight)][sl]
            weight_field += data[(ftype, weight)][sl]

        # Now some fancy footwork
        new_field2 = data.ds.arr(np.zeros((nx, ny, nz)), 
                                 data[(ftype, basename)].units)
        new_field2[1:-1, 1:-1, 1:-1] = new_field / weight_field
        return new_field2

    registry.add_field((ftype, "averaged_%s" % basename), sampling_type="cell", 
                       function=_averaged_field,
                       units=field_units,
                       validators=validators)



from yt.convenience import load

DEFAULT_ATTRS = ("dimensionality", "refine_by", "domain_dimensions",
                 "current_time", "domain_left_edge", "domain_right_edge",
                 "unique_identifier", "current_redshift",
                 "cosmological_simulation", "omega_matter", "omega_lambda",
                 "hubble_constant", "dataset_type")


def get_metadata(path, full_output=False, attrs=DEFAULT_ATTRS):
    ds = load(path)
    metadata = {'filename': path}
    for a in attrs:
        v = getattr(ds, a, None)
        if v is None:
            continue
        if hasattr(v, "tolist"):
            v = v.tolist()
        metadata[a] = v
    if full_output:
        params = {}
        for p, v in ds.parameters.items():
            if hasattr(v, "tolist"):
                v = v.tolist()
            params[p] = v
        metadata['params'] = params
    ds.close()
    return metadata



"""
Tests for making unstructured mesh slices

"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------
import os
import os.path
import tempfile
import shutil
import numpy as np
import yt
from yt.testing import fake_tetrahedral_ds
from yt.testing import fake_hexahedral_ds, small_fake_hexahedral_ds
from yt.utilities.answer_testing.framework import \
    requires_ds, \
    data_dir_load, \
    GenericImageTest
from yt.utilities.lib.geometry_utils import triangle_plane_intersect
from yt.utilities.lib.mesh_triangulation import triangulate_indices


def setup():
    """Test specific setup."""
    from yt.config import ytcfg
    ytcfg["yt", "__withintesting"] = "True"

def compare(ds, field, test_prefix, decimals=12):
    def slice_image(filename_prefix):
        sl = yt.SlicePlot(ds, 'z', field)
        sl.set_log('all', False)
        image_file = sl.save(filename_prefix)
        return image_file

    slice_image.__name__ = "slice_{}".format(test_prefix)
    test = GenericImageTest(ds, slice_image, decimals)
    test.prefix = test_prefix
    return test

tri2 = "SecondOrderTris/RZ_p_no_parts_do_nothing_bcs_cone_out.e"

@requires_ds(tri2)
def test_tri2():
    ds = data_dir_load(tri2, kwargs={'step':-1})
    for field in ds.field_list:
        yield compare(ds, field, "answers_tri2_%s_%s" % (field[0], field[1]))

quad2 = "SecondOrderQuads/lid_driven_out.e"

@requires_ds(quad2)
def test_quad2():
    ds = data_dir_load(quad2, kwargs={'step':-1})
    field_list = [('all', 'T'), ('all', 'vel_x'), ('all', 'vel_y')]
    for field in field_list:
        yield compare(ds, field, "answers_quad2_%s_%s" % (field[0], field[1]))

multi_region = "MultiRegion/two_region_example_out.e"

@requires_ds(multi_region)
def test_multi_region():
    ds = data_dir_load(multi_region, kwargs={'step':-1})
    for field in ds.field_list:
        yield compare(ds, field, "answers_multi_region_%s_%s" % (field[0], field[1]))

def test_mesh_slices():
    # Perform I/O in safe place instead of yt main dir
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    np.random.seed(0x4d3d3d3)

    # tetrahedral ds
    ds = fake_tetrahedral_ds()

    for field in ds.field_list:
        for idir in [0, 1, 2]:
            sl = yt.SlicePlot(ds, idir, field)
            sl.annotate_mesh_lines()
            sl.save()

    # hexahedral ds
    ds = fake_hexahedral_ds()

    for field in ds.field_list:
        for idir in [0, 1, 2]:
            sl = yt.SlicePlot(ds, idir, field)
            sl.annotate_mesh_lines()
            sl.save()

    os.chdir(curdir)
    # clean up
    shutil.rmtree(tmpdir)

def test_perfect_element_intersection():
    # This test tests mesh line annotation where a z=0 slice
    # perfectly intersects the top of a hexahedral element with node
    # z-coordinates containing both -0 and +0. Before
    # https://github.com/yt-project/yt/pull/1437 this test falsely
    # yielded three annotation lines, whereas the correct result is four
    # corresponding to the four edges of the top hex face.

    ds = small_fake_hexahedral_ds()
    indices = ds.index.meshes[0].connectivity_indices
    coords = ds.index.meshes[0].connectivity_coords
    tri_indices = triangulate_indices(indices)
    tri_coords = coords[tri_indices]
    lines = triangle_plane_intersect(2, 0, tri_coords)
    non_zero_lines = 0
    for i in range(lines.shape[0]):
        norm = np.linalg.norm(lines[i][0] - lines[i][1])
        if norm > 1e-8:
            non_zero_lines += 1
    np.testing.assert_equal(non_zero_lines, 4)



"""
ARTIO-specific IO




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------
import numpy as np

from yt.utilities.io_handler import \
    BaseIOHandler


class IOHandlerARTIO(BaseIOHandler):
    _dataset_type = "artio"

    def _read_fluid_selection(self, chunks, selector, fields):
        tr = dict((ftuple, np.empty(0, dtype='float64')) for ftuple in fields)
        cp = 0
        for onechunk in chunks:
            for artchunk in onechunk.objs:
                rv = artchunk.fill(fields, selector)
                for f in fields:
                    tr[f].resize(cp+artchunk.data_size)
                    tr[f][cp:cp+artchunk.data_size] = rv.pop(f)
                cp += artchunk.data_size
        return tr

    def _read_particle_coords(self, chunks, ptf):
        pn = "POSITION_%s"
        chunks = list(chunks)
        fields = [(ptype, pn % ax)
                  for ptype, field_list in ptf.items()
                  for ax in 'XYZ']
        for chunk in chunks: # These should be organized by grid filename
            for subset in chunk.objs:
                rv = dict(**subset.fill_particles(fields))
                for ptype, field_list in sorted(ptf.items()):
                    x, y, z = (np.asarray(rv[ptype][pn % ax], dtype="=f8")
                               for ax in 'XYZ')
                    yield ptype, (x, y, z)
                    rv.pop(ptype)

    def _read_particle_fields(self, chunks, ptf, selector):
        pn = "POSITION_%s"
        chunks = list(chunks)
        fields = [(ptype, fname) for ptype, field_list in ptf.items()
                                 for fname in field_list]
        for ptype, field_list in sorted(ptf.items()):
            for ax in 'XYZ':
                if pn % ax not in field_list:
                    fields.append((ptype, pn % ax))
        for chunk in chunks: # These should be organized by grid filename
            for subset in chunk.objs:
                rv = dict(**subset.fill_particles(fields))
                for ptype, field_list in sorted(ptf.items()):
                    x, y, z = (np.asarray(rv[ptype][pn % ax], dtype="=f8")
                               for ax in 'XYZ')
                    mask = selector.select_points(x, y, z, 0.0)
                    if mask is None: continue
                    for field in field_list:
                        data = np.asarray(rv[ptype][field], "=f8")
                        yield (ptype, field), data[mask]
                    rv.pop(ptype)



# encoding: utf-8
"""
Shader and ShaderProgram wrapper classes for vertex and fragment shaders used 
in Interactive Data Visualization
"""

# ----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# ----------------------------------------------------------------------------

# This is a part of the experimental Interactive Data Visualization 

import os
import OpenGL.GL as GL
import contextlib
from yt.extern.six import add_metaclass
from collections import OrderedDict
from yt.utilities.exceptions import \
    YTInvalidShaderType, \
    YTUnknownUniformKind, \
    YTUnknownUniformSize
from yt.units.yt_array import YTQuantity

known_shaders = {}

class ShaderProgram(object):
    '''
    Wrapper class that compiles and links vertex and fragment shaders
    into a shader program.

    Parameters
    ----------

    vertex_shader : string or :class:`yt.visualization.volume_rendering.shader_objects.VertexShader`
        The vertex shader used in the Interactive Data Visualization pipeline.

    fragment_shader : string or :class:`yt.visualization.volume_rendering.shader_objects.FragmentShader`
        The fragment shader used in the Interactive Data Visualization pipeline.
    '''

    def __init__(self, vertex_shader=None, fragment_shader=None):
        # Don't allow just one.  Either neither or both.
        if vertex_shader is None and fragment_shader is None:
            pass
        elif None not in (vertex_shader, fragment_shader):
            self.link(vertex_shader, fragment_shader)
        else:
            raise RuntimeError
        self._uniform_funcs = OrderedDict()

    def link(self, vertex_shader, fragment_shader):
        # There are more types of shaders, but for now we only allow v&f.
        self.program = GL.glCreateProgram()
        if not isinstance(vertex_shader, VertexShader):
            vertex_shader = VertexShader(vertex_shader)
        if not isinstance(fragment_shader, FragmentShader):
            fragment_shader = FragmentShader(fragment_shader)
        self.vertex_shader = vertex_shader
        self.fragment_shader = fragment_shader
        GL.glAttachShader(self.program, vertex_shader.shader)
        GL.glAttachShader(self.program, fragment_shader.shader)
        GL.glLinkProgram(self.program)
        result = GL.glGetProgramiv(self.program, GL.GL_LINK_STATUS)
        if not result:
            raise RuntimeError(GL.glGetProgramInfoLog(self.program))
        vertex_shader.delete_shader()
        fragment_shader.delete_shader()

    def delete_program(self):
        if self.program is not None:
            GL.glDeleteProgram(self.program)
            self.program = None

    def _guess_uniform_func(self, value):
        # We make a best-effort guess.
        # This does NOT work with arrays of uniforms.
        # First look at the dtype kind.  Fortunately, this falls into either
        # 'f' or 'i', which matches nicely with OpenGL.
        # Note that in some implementations, it seems there is also a 'd' type,
        # but we will not be using that here.
        if isinstance(value, int):
            return GL.glUniform1i
        elif isinstance(value, (YTQuantity, float)):
            return GL.glUniform1f
        else:
            kind = value.dtype.kind
        if kind not in 'if':
            raise YTUnknownUniformKind(kind)
        if len(value.shape) == 1:
            if value.size > 4:
                raise YTUnknownUniformSize(value.size)
            func = self._set_scalar_uniform(kind, value.size)
        elif len(value.shape) == 2:
            if value.shape[0] != value.shape[1]:
                raise YTUnknownUniformSize(value.shape)
            func = self._set_matrix_uniform(kind, value.shape)
        else:
            raise YTUnknownUniformSize(value.shape)
        return func

    def _set_scalar_uniform(self, kind, size_spec):
        gl_func = getattr(GL, "glUniform%s%sv" % (size_spec, kind))
        def _func(location, value):
            return gl_func(location, 1, value)
        return _func

    def _set_matrix_uniform(self, kind, size_spec):
        assert(size_spec[0] == size_spec[1])
        gl_func = getattr(GL, "glUniformMatrix%s%sv" % (size_spec[0], kind))
        def _func(location, value):
            return gl_func(location, 1, GL.GL_TRUE, value)
        return _func

    def _set_uniform(self, name, value):
        # We need to figure out how to pass it in.
        if name not in self._uniform_funcs:
            self._uniform_funcs[name] = self._guess_uniform_func(value)
        loc = GL.glGetUniformLocation(self.program, name)
        return self._uniform_funcs[name](loc, value)

    @contextlib.contextmanager
    def enable(self):
        GL.glUseProgram(self.program)
        yield
        GL.glUseProgram(0)

    def bind_vert_attrib(self, name, bind_loc, size):
        loc = GL.glGetAttribLocation(self.program, name)
        GL.glEnableVertexAttribArray(loc)
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, bind_loc)
        GL.glVertexAttribPointer(loc, size, GL.GL_FLOAT, False, 0, None)

    def disable_vert_attrib(self, name):
        loc = GL.glGetAttribLocation(self.program, name)
        GL.glDisableVertexAttribArray(loc)

class RegisteredShader(type):
    def __init__(cls, name, b, d):
        type.__init__(cls, name, b, d)
        if getattr(cls, "_shader_name", None) is not None:
            known_shaders[cls._shader_name] = cls

@add_metaclass(RegisteredShader)
class Shader(object):
    '''
    Creates a shader from source

    Parameters
    ----------

    source : str
        This can either be a string containing a full source of a shader,
        an absolute path to a source file or a filename of a shader
        residing in the ./shaders/ directory.

    '''

    _shader = None
    _source = None
    _shader_name = None

    def __init__(self, source=None):
        if source:
            self.compile(source)

    def _get_source(self, source):
        if ";" in source:
            # This is probably safe, right?  Enh, probably.
            return source
        if os.path.isfile(source):
            sh_directory = ''
        else:
            sh_directory = os.path.join(os.path.dirname(__file__), "shaders")
        fn = os.path.join(sh_directory, source)
        if not os.path.isfile(fn):
            raise YTInvalidShaderType(source)
        return open(fn, 'r').read()

    def compile(self, source = None, parameters = None):
        if source is None:
            source = self._source
            if source is None: raise RuntimeError
        if parameters is not None:
            raise NotImplementedError
        source = self._get_source(source)
        shader_type_enum = getattr(GL,
            'GL_%s_SHADER' % self.shader_type.upper())
        shader = GL.glCreateShader(shader_type_enum)
        # We could do templating here if we wanted.
        self.shader_source = source
        GL.glShaderSource(shader, source)
        GL.glCompileShader(shader)
        result = GL.glGetShaderiv(shader, GL.GL_COMPILE_STATUS)
        if not(result):
            raise RuntimeError(GL.glGetShaderInfoLog(shader))
        self._shader = shader

    @property
    def shader(self):
        if self._shader is None:
            self.compile()
        return self._shader

    def delete_shader(self):
        if self.shader is not None:
            GL.glDeleteShader(self.shader)
            self._shader = None

    def __del__(self):
        # This is not guaranteed to be called
        self.delete_shader()

class FragmentShader(Shader):
    '''Wrapper class for fragment shaders'''
    shader_type = "fragment"

class VertexShader(Shader):
    '''Wrapper class for vertex shaders'''
    shader_type = "vertex"

class ApplyColormapFragmentShader(FragmentShader):
    '''A second pass fragment shader used to apply a colormap to the result of
    the first pass rendering'''
    _source = "apply_colormap.fragmentshader"
    _shader_name = "apply_colormap.f"

class MaxIntensityFragmentShader(FragmentShader):
    '''A first pass fragment shader that computes Maximum Intensity Projection
    of the data. See :ref:`projection-types` for more information.'''
    _source = "max_intensity.fragmentshader"
    _shader_name = "max_intensity.f"

class NoOpFragmentShader(FragmentShader):
    '''A second pass fragment shader that performs no operation. Usually used if
    the first pass already took care of applying proper color to the data'''
    _source = "noop.fragmentshader"
    _shader_name = "noop.f"

class PassthroughFragmentShader(FragmentShader):
    '''A first pass fragment shader that performs no operation. Used for debug
    purposes. It's distinct from NoOpFragmentShader, because of the number of
    uniforms'''
    _source = "passthrough.fragmentshader"
    _shader_name = "passthrough.f"

class ProjectionFragmentShader(FragmentShader):
    '''A first pass fragment shader that performs unweighted integration of the
    data along the line of sight. See :ref:`projection-types` for more
    information.'''
    _source = "projection.fragmentshader"
    _shader_name = "projection.f"

class TransferFunctionFragmentShader(FragmentShader):
    '''A first pass fragment shader that performs ray casting using transfer
    function. See :ref:`volume-rendering-method` for more details.'''
    _source = "transfer_function.fragmentshader"
    _shader_name = "transfer_function.f"

class DefaultVertexShader(VertexShader):
    '''A first pass vertex shader that translates the location of vertices from
    the world coordinates to the viewing plane coordinates'''
    _source = "default.vertexshader"
    _shader_name = "default.v"

class PassthroughVertexShader(VertexShader):
    '''A second pass vertex shader that performs no operations on vertices'''
    _source = "passthrough.vertexshader"
    _shader_name = "passthrough.v"

class MeshVertexShader(VertexShader):
    '''A vertex shader used for unstructured mesh rendering.'''
    _source = "mesh.vertexshader"
    _shader_name = "mesh.v"

class MeshFragmentShader(FragmentShader):
    '''A vertex shader used for unstructured mesh rendering.'''
    _source = "mesh.fragmentshader"
    _shader_name = "mesh.f"



"""
Cosmology related fields.




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2014, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from .derived_field import \
    ValidateParameter
from .field_exceptions import \
    NeedsConfiguration, \
    NeedsParameter
from .field_plugin_registry import \
    register_field_plugin

from yt.utilities.physical_constants import \
    speed_of_light_cgs

@register_field_plugin
def setup_cosmology_fields(registry, ftype = "gas", slice_info = None):
    unit_system = registry.ds.unit_system
    # slice_info would be the left, the right, and the factor.
    # For example, with the old Enzo-ZEUS fields, this would be:
    # slice(None, -2, None)
    # slice(1, -1, None)
    # 1.0
    # Otherwise, we default to a centered difference.
    if slice_info is None:
        sl_left = slice(None, -2, None)
        sl_right = slice(2, None, None)
        div_fac = 2.0
    else:
        sl_left, sl_right, div_fac = slice_info

    def _matter_density(field, data):
        return data[ftype, "density"] + \
          data[ftype, "dark_matter_density"]

    registry.add_field((ftype, "matter_density"), sampling_type="cell", 
                       function=_matter_density,
                       units=unit_system["density"])

    def _matter_mass(field, data):
        return data[ftype, "matter_density"] * data["index", "cell_volume"]

    registry.add_field((ftype, "matter_mass"), sampling_type="cell", 
                       function=_matter_mass,
                       units=unit_system["mass"])

    # rho_total / rho_cr(z).
    def _overdensity(field, data):
        if not hasattr(data.ds, "cosmological_simulation") or \
          not data.ds.cosmological_simulation:
            raise NeedsConfiguration("cosmological_simulation", 1)
        co = data.ds.cosmology
        return data[ftype, "matter_density"] / \
          co.critical_density(data.ds.current_redshift)

    registry.add_field((ftype, "overdensity"), sampling_type="cell", 
                       function=_overdensity,
                       units="")

    # rho_baryon / <rho_baryon>
    def _baryon_overdensity(field, data):
        if not hasattr(data.ds, "cosmological_simulation") or \
          not data.ds.cosmological_simulation:
            raise NeedsConfiguration("cosmological_simulation", 1)
        omega_baryon = data.get_field_parameter("omega_baryon")
        if omega_baryon is None:
            raise NeedsParameter("omega_baryon")
        co = data.ds.cosmology
        # critical_density(z) ~ omega_lambda + omega_matter * (1 + z)^3
        # mean matter density(z) ~ omega_matter * (1 + z)^3
        return data[ftype, "density"] / omega_baryon / co.critical_density(0.0) / \
          (1.0 + data.ds.current_redshift)**3

    registry.add_field((ftype, "baryon_overdensity"), sampling_type="cell", 
                       function=_baryon_overdensity,
                       units="",
                       validators=[ValidateParameter("omega_baryon")])

    # rho_matter / <rho_matter>
    def _matter_overdensity(field, data):
        if not hasattr(data.ds, "cosmological_simulation") or \
          not data.ds.cosmological_simulation:
            raise NeedsConfiguration("cosmological_simulation", 1)
        co = data.ds.cosmology
        # critical_density(z) ~ omega_lambda + omega_matter * (1 + z)^3
        # mean density(z) ~ omega_matter * (1 + z)^3
        return data[ftype, "matter_density"] / data.ds.omega_matter / \
          co.critical_density(0.0) / \
          (1.0 + data.ds.current_redshift)**3

    registry.add_field((ftype, "matter_overdensity"), sampling_type="cell", 
                       function=_matter_overdensity,
                       units="")

    # r / r_vir
    def _virial_radius_fraction(field, data):
        virial_radius = data.get_field_parameter("virial_radius")
        if virial_radius == 0.0:
            ret = 0.0
        else:
            ret = data["radius"] / virial_radius
        return ret

    registry.add_field(("index", "virial_radius_fraction"), sampling_type="cell", 
                       function=_virial_radius_fraction,
                       validators=[ValidateParameter("virial_radius")],
                       units="")

    # Weak lensing convergence.
    # Eqn 4 of Metzler, White, & Loken (2001, ApJ, 547, 560).
    # This needs to be checked for accuracy.
    def _weak_lensing_convergence(field, data):
        if not hasattr(data.ds, "cosmological_simulation") or \
          not data.ds.cosmological_simulation:
            raise NeedsConfiguration("cosmological_simulation", 1)
        co = data.ds.cosmology
        observer_redshift = data.get_field_parameter('observer_redshift')
        source_redshift = data.get_field_parameter('source_redshift')

        # observer to lens
        dl = co.angular_diameter_distance(observer_redshift, data.ds.current_redshift)
        # observer to source
        ds = co.angular_diameter_distance(observer_redshift, source_redshift)
        # lens to source
        dls = co.angular_diameter_distance(data.ds.current_redshift, source_redshift)

        # removed the factor of 1 / a to account for the fact that we are projecting
        # with a proper distance.
        return (1.5 * (co.hubble_constant / speed_of_light_cgs)**2 * (dl * dls / ds) * \
          data[ftype, "matter_overdensity"]).in_units("1/cm")

    registry.add_field((ftype, "weak_lensing_convergence"), sampling_type="cell", 
                       function=_weak_lensing_convergence,
                       units=unit_system["length"]**-1,
        validators=[ValidateParameter("observer_redshift"),
                    ValidateParameter("source_redshift")])



from mpi4py import MPI
import os
import sys
import yt
from yt.analysis_modules.halo_analysis.api import \
    HaloCatalog
from yt.data_objects.particle_filters import \
    particle_filter
yt.enable_parallelism()

method = sys.argv[1]
comm = MPI.Comm.Get_parent()

methods = {"fof": {}, "hop": {},
           "rockstar": {"num_readers":1,
                        "num_writers":1,
                        "particle_type":"dark_matter"}}

@particle_filter("dark_matter", requires=["creation_time"])
def _dm_filter(pfilter, data):
    return data["creation_time"] <= 0.0

ds = yt.load("Enzo_64/DD0043/data0043")
ds.add_particle_filter("dark_matter")

output_dir = os.path.join(os.path.dirname(__file__),
                          "halo_catalogs", method)
hc = HaloCatalog(data_ds=ds, output_dir=output_dir,
                 finder_method=method, finder_kwargs=methods[method])
hc.create()

comm.Disconnect()



from yt.funcs import issue_deprecation_warning

issue_deprecation_warning("The fits_image classes have been moved to yt.visualization "
                          "and can be imported from yt directly. yt.utilities.fits_image "
                          "is deprecated.")

from yt.visualization.fits_image import \
    FITSImageData, FITSSlice, FITSProjection, \
    FITSOffAxisSlice, FITSOffAxisProjection


"""
Profile classes, to deal with generating and obtaining profiles



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from yt.fields.derived_field import DerivedField
from yt.frontends.ytdata.utilities import \
    save_as_dataset
from yt.funcs import \
    get_output_filename, \
    ensure_list, \
    iterable, \
    issue_deprecation_warning, \
    mylog
from yt.units.yt_array import \
    array_like_field, \
    YTQuantity
from yt.units.unit_object import Unit
from yt.data_objects.field_data import YTFieldData
from yt.utilities.exceptions import \
    YTIllDefinedProfile
from yt.utilities.lib.misc_utilities import \
    new_bin_profile1d, \
    new_bin_profile2d, \
    new_bin_profile3d
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    ParallelAnalysisInterface, parallel_objects
from yt.utilities.lib.particle_mesh_operations import \
    CICDeposit_2, \
    NGPDeposit_2


def _sanitize_min_max_units(amin, amax, finfo, registry):
    # returns a copy of amin and amax, converted to finfo's output units
    umin = getattr(amin, 'units', None)
    umax = getattr(amax, 'units', None)
    if umin is None:
        umin = Unit(finfo.output_units, registry=registry)
        rmin = YTQuantity(amin, umin)
    else:
        rmin = amin.in_units(finfo.output_units)
    if umax is None:
        umax = Unit(finfo.output_units, registry=registry)
        rmax = YTQuantity(amax, umax)
    else:
        rmax = amax.in_units(finfo.output_units)
    return rmin, rmax

def preserve_source_parameters(func):
    def save_state(*args, **kwargs):
        # Temporarily replace the 'field_parameters' for a
        # grid with the 'field_parameters' for the data source
        prof = args[0]
        source = args[1]
        if hasattr(source, 'field_parameters'):
            old_params = source.field_parameters
            source.field_parameters = prof._data_source.field_parameters
            tr = func(*args, **kwargs)
            source.field_parameters = old_params
        else:
            tr = func(*args, **kwargs)
        return tr
    return save_state


class ProfileFieldAccumulator(object):
    def __init__(self, n_fields, size):
        shape = size + (n_fields,)
        self.values = np.zeros(shape, dtype="float64")
        self.mvalues = np.zeros(shape, dtype="float64")
        self.qvalues = np.zeros(shape, dtype="float64")
        self.used = np.zeros(size, dtype='bool')
        self.weight_values = np.zeros(size, dtype="float64")

class ProfileND(ParallelAnalysisInterface):
    """The profile object class"""
    def __init__(self, data_source, weight_field = None):
        self.data_source = data_source
        self.ds = data_source.ds
        self.field_map = {}
        self.field_info = {}
        self.field_data = YTFieldData()
        if weight_field is not None:
            self.standard_deviation = YTFieldData()
            weight_field = self.data_source._determine_fields(weight_field)[0]
        else:
            self.standard_deviation = None
        self.weight_field = weight_field
        self.field_units = {}
        ParallelAnalysisInterface.__init__(self, comm=data_source.comm)

    @property
    def variance(self):
        issue_deprecation_warning("""
profile.variance incorrectly returns the profile standard deviation and has 
been deprecated, use profile.standard_deviation instead."""
        )
        return self.standard_deviation

    def add_fields(self, fields):
        """Add fields to profile

        Parameters
        ----------
        fields : list of field names
            A list of fields to create profile histograms for
        
        """
        fields = self.data_source._determine_fields(fields)
        for f in fields:
            self.field_info[f] = self.data_source.ds.field_info[f]
        temp_storage = ProfileFieldAccumulator(len(fields), self.size)
        citer = self.data_source.chunks([], "io")
        for chunk in parallel_objects(citer):
            self._bin_chunk(chunk, fields, temp_storage)
        self._finalize_storage(fields, temp_storage)

    def set_field_unit(self, field, new_unit):
        """Sets a new unit for the requested field

        Parameters
        ----------
        field : string or field tuple
           The name of the field that is to be changed.

        new_unit : string or Unit object
           The name of the new unit.
        """
        if field in self.field_units:
            self.field_units[field] = \
                Unit(new_unit, registry=self.ds.unit_registry)
        else:
            fd = self.field_map[field]
            if fd in self.field_units:
                self.field_units[fd] = \
                    Unit(new_unit, registry=self.ds.unit_registry)
            else:
                raise KeyError("%s not in profile!" % (field))

    def _finalize_storage(self, fields, temp_storage):
        # We use our main comm here
        # This also will fill _field_data

        for i, field in enumerate(fields):
            # q values are returned as q * weight but we want just q
            temp_storage.qvalues[..., i][temp_storage.used] /= \
              temp_storage.weight_values[temp_storage.used]

        # get the profile data from all procs
        all_store = {self.comm.rank: temp_storage}
        all_store = self.comm.par_combine_object(all_store,
                                                 "join", datatype="dict")

        all_val = np.zeros_like(temp_storage.values)
        all_mean = np.zeros_like(temp_storage.mvalues)
        all_std = np.zeros_like(temp_storage.qvalues)
        all_weight = np.zeros_like(temp_storage.weight_values)
        all_used = np.zeros_like(temp_storage.used, dtype="bool")

        # Combine the weighted mean and standard deviation from each processor.
        # For two samples with total weight, mean, and standard deviation 
        # given by w, m, and s, their combined mean and standard deviation are:
        # m12 = (m1 * w1 + m2 * w2) / (w1 + w2)
        # s12 = (m1 * (s1**2 + (m1 - m12)**2) + 
        #        m2 * (s2**2 + (m2 - m12)**2)) / (w1 + w2)
        # Here, the mvalues are m and the qvalues are s**2.
        for p in sorted(all_store.keys()):
            all_used += all_store[p].used
            old_mean = all_mean.copy()
            old_weight = all_weight.copy()
            all_weight[all_store[p].used] += \
              all_store[p].weight_values[all_store[p].used]
            for i, field in enumerate(fields):
                all_val[..., i][all_store[p].used] += \
                  all_store[p].values[..., i][all_store[p].used]

                all_mean[..., i][all_store[p].used] = \
                  (all_mean[..., i] * old_weight +
                   all_store[p].mvalues[..., i] *
                   all_store[p].weight_values)[all_store[p].used] / \
                   all_weight[all_store[p].used]

                all_std[..., i][all_store[p].used] = \
                  (old_weight * (all_std[..., i] +
                                 (old_mean[..., i] - all_mean[..., i])**2) +
                   all_store[p].weight_values *
                   (all_store[p].qvalues[..., i] + 
                    (all_store[p].mvalues[..., i] -
                     all_mean[..., i])**2))[all_store[p].used] / \
                    all_weight[all_store[p].used]

        all_std = np.sqrt(all_std)
        del all_store
        self.used = all_used
        blank = ~all_used

        self.weight = all_weight
        self.weight[blank] = 0.0

        for i, field in enumerate(fields):
            if self.weight_field is None:
                self.field_data[field] = \
                  array_like_field(self.data_source, 
                                   all_val[...,i], field)
            else:
                self.field_data[field] = \
                  array_like_field(self.data_source, 
                                   all_mean[...,i], field)
                self.standard_deviation[field] = \
                  array_like_field(self.data_source,
                                   all_std[...,i], field)
                self.standard_deviation[field][blank] = 0.0
            self.field_data[field][blank] = 0.0
            self.field_units[field] = self.field_data[field].units
            if isinstance(field, tuple):
                self.field_map[field[1]] = field
            else:
                self.field_map[field] = field

    def _bin_chunk(self, chunk, fields, storage):
        raise NotImplementedError

    def _filter(self, bin_fields):
        # cut_points is set to be everything initially, but
        # we also want to apply a filtering based on min/max
        filter = np.ones(bin_fields[0].shape, dtype='bool')
        for (mi, ma), data in zip(self.bounds, bin_fields):
            filter &= (data > mi)
            filter &= (data < ma)
        return filter, [data[filter] for data in bin_fields]

    def _get_data(self, chunk, fields):
        # We are using chunks now, which will manage the field parameters and
        # the like.
        bin_fields = [chunk[bf] for bf in self.bin_fields]
        # We want to make sure that our fields are within the bounds of the
        # binning
        filter, bin_fields = self._filter(bin_fields)
        if not np.any(filter): return None
        arr = np.zeros((bin_fields[0].size, len(fields)), dtype="float64")
        for i, field in enumerate(fields):
            units = chunk.ds.field_info[field].output_units
            arr[:,i] = chunk[field][filter].in_units(units)
        if self.weight_field is not None:
            units = chunk.ds.field_info[self.weight_field].output_units
            weight_data = chunk[self.weight_field].in_units(units)
        else:
            weight_data = np.ones(filter.shape, dtype="float64")
        weight_data = weight_data[filter]
        # So that we can pass these into
        return arr, weight_data, bin_fields

    def __getitem__(self, field):
        fname = self.field_map.get(field, None)
        if fname is None:
            if isinstance(field, tuple):
                fname = self.field_map.get(field[1], None)
            elif isinstance(field, DerivedField):
                fname = self.field_map.get(field.name[1], None)
        if fname is None:
            raise KeyError(field)
        else:
            if getattr(self, 'fractional', False):
                return self.field_data[fname]
            else:
                return self.field_data[fname].in_units(self.field_units[fname])

    def items(self):
        return [(k,self[k]) for k in self.field_data.keys()]

    def keys(self):
        return self.field_data.keys()

    def __iter__(self):
        return sorted(self.items())

    def _get_bins(self, mi, ma, n, take_log):
        if take_log:
            ret = np.logspace(np.log10(mi), np.log10(ma), n+1)
            # at this point ret[0] and ret[-1] are not exactly equal to
            # mi and ma due to round-off error. Let's force them to be
            # mi and ma exactly to avoid incorrectly discarding cells near
            # the edges. See Issue #1300.
            ret[0], ret[-1] = mi, ma
            return ret
        else:
            return np.linspace(mi, ma, n+1)

    def save_as_dataset(self, filename=None):
        r"""Export a profile to a reloadable yt dataset.

        This function will take a profile and output a dataset
        containing all relevant fields.  The resulting dataset
        can be reloaded as a yt dataset.

        Parameters
        ----------
        filename : str, optional
            The name of the file to be written.  If None, the name
            will be a combination of the original dataset plus
            the type of object, e.g., Profile1D.

        Returns
        -------
        filename : str
            The name of the file that has been created.

        Examples
        --------

        >>> import yt
        >>> ds = yt.load("enzo_tiny_cosmology/DD0046/DD0046")
        >>> ad = ds.all_data()
        >>> profile = yt.create_profile(ad, ["density", "temperature"],
        ...                            "cell_mass", weight_field=None,
        ...                             n_bins=(128, 128))
        >>> fn = profile.save_as_dataset()
        >>> prof_ds = yt.load(fn)
        >>> print (prof_ds.data["cell_mass"])
        (128, 128)
        >>> print (prof_ds.data["x"].shape) # x bins as 1D array
        (128,)
        >>> print (prof_ds.data["density"]) # x bins as 2D array
        (128, 128)
        >>> p = yt.PhasePlot(prof_ds.data, "density", "temperature",
        ...                  "cell_mass", weight_field=None)
        >>> p.save()

        """

        keyword = "%s_%s" % (str(self.ds), self.__class__.__name__)
        filename = get_output_filename(filename, keyword, ".h5")

        args = ("field", "log")
        extra_attrs = {"data_type": "yt_profile",
                       "profile_dimensions": self.size,
                       "weight_field": self.weight_field,
                       "fractional": self.fractional,
                       "accumulation": self.accumulation}
        data = {}
        data.update(self.field_data)
        data["weight"] = self.weight
        data["used"] = self.used.astype("float64")

        dimensionality = 0
        bin_data = []
        for ax in "xyz":
            if hasattr(self, ax):
                dimensionality += 1
                data[ax] = getattr(self, ax)
                bin_data.append(data[ax])
                bin_field_name = "%s_bins" % ax
                data[bin_field_name] = getattr(self, bin_field_name)
                extra_attrs["%s_range" % ax] = self.ds.arr([data[bin_field_name][0],
                                                            data[bin_field_name][-1]])
                for arg in args:
                    key = "%s_%s" % (ax, arg)
                    extra_attrs[key] = getattr(self, key)

        bin_fields = np.meshgrid(*bin_data)
        for i, ax in enumerate("xyz"[:dimensionality]):
            data[getattr(self, "%s_field" % ax)] = bin_fields[i]

        extra_attrs["dimensionality"] = dimensionality
        ftypes = dict([(field, "data") for field in data])
        save_as_dataset(self.ds, filename, data, field_types=ftypes,
                        extra_attrs=extra_attrs)

        return filename

class ProfileNDFromDataset(ProfileND):
    """
    An ND profile object loaded from a ytdata dataset.
    """
    def __init__(self, ds):
        ProfileND.__init__(self, ds.data, ds.parameters["weight_field"])
        self.fractional = ds.parameters["fractional"]
        self.accumulation = ds.parameters["accumulation"]
        exclude_fields = ["used", "weight"]
        for ax in "xyz"[:ds.dimensionality]:
            setattr(self, ax, ds.data[ax])
            setattr(self, "%s_bins" % ax, ds.data["%s_bins" % ax])
            field_name = tuple(ds.parameters["%s_field" % ax])
            setattr(self, "%s_field" % ax, field_name)
            self.field_info[field_name] = ds.field_info[field_name]
            setattr(self, "%s_log" % ax, ds.parameters["%s_log" % ax])
            exclude_fields.extend([ax, "%s_bins" % ax,
                                   ds.parameters["%s_field" % ax][1]])
        self.weight = ds.data["weight"]
        self.used = ds.data["used"].d.astype(bool)
        profile_fields = [f for f in ds.field_list
                          if f[1] not in exclude_fields]
        for field in profile_fields:
            self.field_map[field[1]] = field
            self.field_data[field] = ds.data[field]
            self.field_info[field] = ds.field_info[field]
            self.field_units[field] = ds.data[field].units

class Profile1D(ProfileND):
    """An object that represents a 1D profile.

    Parameters
    ----------

    data_source : AMD3DData object
        The data object to be profiled
    x_field : string field name
        The field to profile as a function of
    x_n : integer
        The number of bins along the x direction.
    x_min : float
        The minimum value of the x profile field. If supplied without units,
        assumed to be in the output units for x_field.
    x_max : float
        The maximum value of the x profile field. If supplied without units,
        assumed to be in the output units for x_field.
    x_log : boolean
        Controls whether or not the bins for the x field are evenly
        spaced in linear (False) or log (True) space.
    weight_field : string field name
        The field to weight the profiled fields by.

    """
    def __init__(self, data_source, x_field, x_n, x_min, x_max, x_log,
                 weight_field = None):
        super(Profile1D, self).__init__(data_source, weight_field)
        self.x_field = data_source._determine_fields(x_field)[0]
        self.field_info[self.x_field] = \
            self.data_source.ds.field_info[self.x_field]
        self.x_log = x_log
        x_min, x_max = _sanitize_min_max_units(
            x_min, x_max, self.field_info[self.x_field], self.ds.unit_registry)
        self.x_bins = array_like_field(data_source,
                                       self._get_bins(x_min, x_max, x_n, x_log),
                                       self.x_field)
        self.size = (self.x_bins.size - 1,)
        self.bin_fields = (self.x_field,)
        self.x = 0.5*(self.x_bins[1:]+self.x_bins[:-1])

    def _bin_chunk(self, chunk, fields, storage):
        rv = self._get_data(chunk, fields)
        if rv is None: return
        fdata, wdata, (bf_x,) = rv
        bf_x.convert_to_units(self.field_info[self.x_field].output_units)
        bin_ind = np.digitize(bf_x, self.x_bins) - 1
        new_bin_profile1d(bin_ind, wdata, fdata,
                      storage.weight_values, storage.values,
                      storage.mvalues, storage.qvalues,
                      storage.used)

        # We've binned it!

    def set_x_unit(self, new_unit):
        """Sets a new unit for the x field

        parameters
        ----------
        new_unit : string or Unit object
           The name of the new unit.
        """
        self.x_bins.convert_to_units(new_unit)
        self.x = 0.5*(self.x_bins[1:]+self.x_bins[:-1])

    @property
    def bounds(self):
        return ((self.x_bins[0], self.x_bins[-1]),)

    def plot(self):
        r"""
        This returns a :class:~yt.visualization.profile_plotter.ProfilePlot
        with the fields that have been added to this object.
        """
        from yt.visualization.profile_plotter import ProfilePlot
        return ProfilePlot.from_profiles(self)

class Profile1DFromDataset(ProfileNDFromDataset, Profile1D):
    """
    A 1D profile object loaded from a ytdata dataset.
    """

    def __init(self, ds):
        ProfileNDFromDataset.__init__(self, ds)

class Profile2D(ProfileND):
    """An object that represents a 2D profile.

    Parameters
    ----------

    data_source : AMD3DData object
        The data object to be profiled
    x_field : string field name
        The field to profile as a function of along the x axis.
    x_n : integer
        The number of bins along the x direction.
    x_min : float
        The minimum value of the x profile field. If supplied without units,
        assumed to be in the output units for x_field.
    x_max : float
        The maximum value of the x profile field. If supplied without units,
        assumed to be in the output units for x_field.
    x_log : boolean
        Controls whether or not the bins for the x field are evenly
        spaced in linear (False) or log (True) space.
    y_field : string field name
        The field to profile as a function of along the y axis
    y_n : integer
        The number of bins along the y direction.
    y_min : float
        The minimum value of the y profile field. If supplied without units,
        assumed to be in the output units for y_field.
    y_max : float
        The maximum value of the y profile field. If supplied without units,
        assumed to be in the output units for y_field.
    y_log : boolean
        Controls whether or not the bins for the y field are evenly
        spaced in linear (False) or log (True) space.
    weight_field : string field name
        The field to weight the profiled fields by.

    """
    def __init__(self, data_source,
                 x_field, x_n, x_min, x_max, x_log,
                 y_field, y_n, y_min, y_max, y_log,
                 weight_field = None):
        super(Profile2D, self).__init__(data_source, weight_field)
        # X
        self.x_field = data_source._determine_fields(x_field)[0]
        self.x_log = x_log
        self.field_info[self.x_field] = \
            self.data_source.ds.field_info[self.x_field]
        x_min, x_max = _sanitize_min_max_units(
            x_min, x_max, self.field_info[self.x_field], self.ds.unit_registry)
        self.x_bins = array_like_field(data_source,
                                       self._get_bins(x_min, x_max, x_n, x_log),
                                       self.x_field)
        # Y
        self.y_field = data_source._determine_fields(y_field)[0]
        self.y_log = y_log
        self.field_info[self.y_field] = \
            self.data_source.ds.field_info[self.y_field]
        y_min, y_max = _sanitize_min_max_units(
            y_min, y_max, self.field_info[self.y_field], self.ds.unit_registry)
        self.y_bins = array_like_field(data_source,
                                       self._get_bins(y_min, y_max, y_n, y_log),
                                       self.y_field)

        self.size = (self.x_bins.size - 1, self.y_bins.size - 1)

        self.bin_fields = (self.x_field, self.y_field)
        self.x = 0.5*(self.x_bins[1:]+self.x_bins[:-1])
        self.y = 0.5*(self.y_bins[1:]+self.y_bins[:-1])

    def _bin_chunk(self, chunk, fields, storage):
        rv = self._get_data(chunk, fields)
        if rv is None: return
        fdata, wdata, (bf_x, bf_y) = rv
        bf_x.convert_to_units(self.field_info[self.x_field].output_units)
        bin_ind_x = np.digitize(bf_x, self.x_bins) - 1
        bf_y.convert_to_units(self.field_info[self.y_field].output_units)
        bin_ind_y = np.digitize(bf_y, self.y_bins) - 1
        new_bin_profile2d(bin_ind_x, bin_ind_y, wdata, fdata,
                      storage.weight_values, storage.values,
                      storage.mvalues, storage.qvalues,
                      storage.used)
        # We've binned it!

    def set_x_unit(self, new_unit):
        """Sets a new unit for the x field

        parameters
        ----------
        new_unit : string or Unit object
           The name of the new unit.
        """
        self.x_bins.convert_to_units(new_unit)
        self.x = 0.5*(self.x_bins[1:]+self.x_bins[:-1])

    def set_y_unit(self, new_unit):
        """Sets a new unit for the y field

        parameters
        ----------
        new_unit : string or Unit object
           The name of the new unit.
        """
        self.y_bins.convert_to_units(new_unit)
        self.y = 0.5*(self.y_bins[1:]+self.y_bins[:-1])

    @property
    def bounds(self):
        return ((self.x_bins[0], self.x_bins[-1]),
                (self.y_bins[0], self.y_bins[-1]))

    def plot(self):
        r"""
        This returns a :class:~yt.visualization.profile_plotter.PhasePlot with
        the fields that have been added to this object.
        """
        from yt.visualization.profile_plotter import PhasePlot
        return PhasePlot.from_profile(self)

class Profile2DFromDataset(ProfileNDFromDataset, Profile2D):
    """
    A 2D profile object loaded from a ytdata dataset.
    """

    def __init(self, ds):
        ProfileNDFromDataset.__init__(self, ds)

class ParticleProfile(Profile2D):
    """An object that represents a *deposited* 2D profile. This is like a
    Profile2D, except that it is intended for particle data. Instead of just
    binning the particles, the added fields will be deposited onto the mesh
    using either the nearest-grid-point or cloud-in-cell interpolation kernels.

    Parameters
    ----------

    data_source : AMD3DData object
        The data object to be profiled
    x_field : string field name
        The field to profile as a function of along the x axis.
    x_n : integer
        The number of bins along the x direction.
    x_min : float
        The minimum value of the x profile field. If supplied without units,
        assumed to be in the output units for x_field.
    x_max : float
        The maximum value of the x profile field. If supplied without units,
        assumed to be in the output units for x_field.
    y_field : string field name
        The field to profile as a function of along the y axis
    y_n : integer
        The number of bins along the y direction.
    y_min : float
        The minimum value of the y profile field. If supplied without units,
        assumed to be in the output units for y_field.
    y_max : float
        The maximum value of the y profile field. If supplied without units,
        assumed to be in the output units for y_field.
    weight_field : string field name
        The field to use for weighting. Default is None.
    deposition : string, optional
        The interpolation kernal to be used for
        deposition. Valid choices:
        "ngp" : nearest grid point interpolation
        "cic" : cloud-in-cell interpolation

    """
    accumulation = False
    fractional = False

    def __init__(self, data_source,
                 x_field, x_n, x_min, x_max, x_log,
                 y_field, y_n, y_min, y_max, y_log,
                 weight_field=None, deposition="ngp"):

        x_field = data_source._determine_fields(x_field)[0]
        y_field = data_source._determine_fields(y_field)[0]

        if deposition not in ['ngp', 'cic']:
            raise NotImplementedError(deposition)
        elif (x_log or y_log) and deposition != 'ngp':
            mylog.warning('cic deposition is only supported for linear axis '
                          'scales, falling back to ngp deposition')
            deposition = 'ngp'

        self.deposition = deposition

        # set the log parameters to False (since that doesn't make much sense
        # for deposited data) and also turn off the weight field.
        super(ParticleProfile, self).__init__(data_source,
                                              x_field,
                                              x_n, x_min, x_max, x_log,
                                              y_field,
                                              y_n, y_min, y_max, y_log,
                                              weight_field=weight_field)

    # Either stick the particle field in the nearest bin,
    # or spread it out using the 2D CIC deposition function
    def _bin_chunk(self, chunk, fields, storage):
        rv = self._get_data(chunk, fields)
        if rv is None: return
        fdata, wdata, (bf_x, bf_y) = rv
        # make sure everything has the same units before deposition.
        # the units will be scaled to the correct values later.

        if self.deposition == "ngp":
            func = NGPDeposit_2
        elif self.deposition == "cic":
            func = CICDeposit_2

        for fi, field in enumerate(fields):
            if self.weight_field is None:
                deposit_vals = fdata[:, fi]
            else:
                deposit_vals = wdata*fdata[:, fi]

            func(bf_x, bf_y, deposit_vals, fdata[:, fi].size,
                 storage.values[:, :, fi], self.x_bins, self.y_bins)

            locs = storage.values[:, :, fi] != 0.0
            storage.used[locs] = True

            if self.weight_field is not None:
                func(bf_x, bf_y, wdata, fdata[:, fi].size,
                     storage.weight_values, self.x_bins, self.y_bins)
            else:
                storage.weight_values[locs] = 1.0
            storage.mvalues[locs, fi] = \
                storage.values[locs, fi] / storage.weight_values[locs]
        # We've binned it!


class Profile3D(ProfileND):
    """An object that represents a 2D profile.

    Parameters
    ----------

    data_source : AMD3DData object
        The data object to be profiled
    x_field : string field name
        The field to profile as a function of along the x axis.
    x_n : integer
        The number of bins along the x direction.
    x_min : float
        The minimum value of the x profile field. If supplied without units,
        assumed to be in the output units for x_field.
    x_max : float
        The maximum value of the x profile field. If supplied without units,
        assumed to be in the output units for x_field.
    x_log : boolean
        Controls whether or not the bins for the x field are evenly
        spaced in linear (False) or log (True) space.
    y_field : string field name
        The field to profile as a function of along the y axis
    y_n : integer
        The number of bins along the y direction.
    y_min : float
        The minimum value of the y profile field. If supplied without units,
        assumed to be in the output units for y_field.
    y_max : float
        The maximum value of the y profile field. If supplied without units,
        assumed to be in the output units for y_field.
    y_log : boolean
        Controls whether or not the bins for the y field are evenly
        spaced in linear (False) or log (True) space.
    z_field : string field name
        The field to profile as a function of along the z axis
    z_n : integer
        The number of bins along the z direction.
    z_min : float
        The minimum value of the z profile field. If supplied without units,
        assumed to be in the output units for z_field.
    z_max : float
        The maximum value of thee z profile field. If supplied without units,
        assumed to be in the output units for z_field.
    z_log : boolean
        Controls whether or not the bins for the z field are evenly
        spaced in linear (False) or log (True) space.
    weight_field : string field name
        The field to weight the profiled fields by.

    """
    def __init__(self, data_source,
                 x_field, x_n, x_min, x_max, x_log,
                 y_field, y_n, y_min, y_max, y_log,
                 z_field, z_n, z_min, z_max, z_log,
                 weight_field = None):
        super(Profile3D, self).__init__(data_source, weight_field)
        # X
        self.x_field = data_source._determine_fields(x_field)[0]
        self.x_log = x_log
        self.field_info[self.x_field] = \
            self.data_source.ds.field_info[self.x_field]
        x_min, x_max = _sanitize_min_max_units(
            x_min, x_max, self.field_info[self.x_field], self.ds.unit_registry)
        self.x_bins = array_like_field(data_source,
                                       self._get_bins(x_min, x_max, x_n, x_log),
                                       self.x_field)
        # Y
        self.y_field = data_source._determine_fields(y_field)[0]
        self.y_log = y_log
        self.field_info[self.y_field] = \
            self.data_source.ds.field_info[self.y_field]
        y_min, y_max = _sanitize_min_max_units(
            y_min, y_max, self.field_info[self.y_field], self.ds.unit_registry)
        self.y_bins = array_like_field(data_source,
                                       self._get_bins(y_min, y_max, y_n, y_log),
                                       self.y_field)
        # Z
        self.z_field = data_source._determine_fields(z_field)[0]
        self.z_log = z_log
        self.field_info[self.z_field] = \
            self.data_source.ds.field_info[self.z_field]
        z_min, z_max = _sanitize_min_max_units(
            z_min, z_max, self.field_info[self.z_field], self.ds.unit_registry)
        self.z_bins = array_like_field(data_source,
                                       self._get_bins(z_min, z_max, z_n, z_log),
                                       self.z_field)

        self.size = (self.x_bins.size - 1,
                     self.y_bins.size - 1,
                     self.z_bins.size - 1)

        self.bin_fields = (self.x_field, self.y_field, self.z_field)
        self.x = 0.5*(self.x_bins[1:]+self.x_bins[:-1])
        self.y = 0.5*(self.y_bins[1:]+self.y_bins[:-1])
        self.z = 0.5*(self.z_bins[1:]+self.z_bins[:-1])

    def _bin_chunk(self, chunk, fields, storage):
        rv = self._get_data(chunk, fields)
        if rv is None: return
        fdata, wdata, (bf_x, bf_y, bf_z) = rv
        bf_x.convert_to_units(self.field_info[self.x_field].output_units)
        bin_ind_x = np.digitize(bf_x, self.x_bins) - 1
        bf_y.convert_to_units(self.field_info[self.y_field].output_units)
        bin_ind_y = np.digitize(bf_y, self.y_bins) - 1
        bf_z.convert_to_units(self.field_info[self.z_field].output_units)
        bin_ind_z = np.digitize(bf_z, self.z_bins) - 1
        new_bin_profile3d(bin_ind_x, bin_ind_y, bin_ind_z, wdata, fdata,
                      storage.weight_values, storage.values,
                      storage.mvalues, storage.qvalues,
                      storage.used)
        # We've binned it!

    @property
    def bounds(self):
        return ((self.x_bins[0], self.x_bins[-1]),
                (self.y_bins[0], self.y_bins[-1]),
                (self.z_bins[0], self.z_bins[-1]))

    def set_x_unit(self, new_unit):
        """Sets a new unit for the x field

        parameters
        ----------
        new_unit : string or Unit object
           The name of the new unit.
        """
        self.x_bins.convert_to_units(new_unit)
        self.x = 0.5*(self.x_bins[1:]+self.x_bins[:-1])

    def set_y_unit(self, new_unit):
        """Sets a new unit for the y field

        parameters
        ----------
        new_unit : string or Unit object
           The name of the new unit.
        """
        self.y_bins.convert_to_units(new_unit)
        self.y = 0.5*(self.y_bins[1:]+self.y_bins[:-1])

    def set_z_unit(self, new_unit):
        """Sets a new unit for the z field

        parameters
        ----------
        new_unit : string or Unit object
           The name of the new unit.
        """
        self.z_bins.convert_to_units(new_unit)
        self.z = 0.5*(self.z_bins[1:]+self.z_bins[:-1])

class Profile3DFromDataset(ProfileNDFromDataset, Profile3D):
    """
    A 2D profile object loaded from a ytdata dataset.
    """

    def __init(self, ds):
        ProfileNDFromDataset.__init__(self, ds)

def sanitize_field_tuple_keys(input_dict, data_source):
    if input_dict is not None:
        dummy = {}
        for item in input_dict:
            dummy[data_source._determine_fields(item)[0]] = input_dict[item]
        return dummy
    else:
        return input_dict


def create_profile(data_source, bin_fields, fields, n_bins=64,
                   extrema=None, logs=None, units=None,
                   weight_field="cell_mass",
                   accumulation=False, fractional=False,
                   deposition='ngp'):
    r"""
    Create a 1, 2, or 3D profile object.

    The dimensionality of the profile object is chosen by the number of
    fields given in the bin_fields argument.

    Parameters
    ----------
    data_source : YTSelectionContainer Object
        The data object to be profiled.
    bin_fields : list of strings
        List of the binning fields for profiling.
    fields : list of strings
        The fields to be profiled.
    n_bins : int or list of ints
        The number of bins in each dimension.  If None, 64 bins for
        each bin are used for each bin field.
        Default: 64.
    extrema : dict of min, max tuples
        Minimum and maximum values of the bin_fields for the profiles.
        The keys correspond to the field names. Defaults to the extrema
        of the bin_fields of the dataset. If a units dict is provided, extrema
        are understood to be in the units specified in the dictionary.
    logs : dict of boolean values
        Whether or not to log the bin_fields for the profiles.
        The keys correspond to the field names. Defaults to the take_log
        attribute of the field.
    units : dict of strings
        The units of the fields in the profiles, including the bin_fields.
    weight_field : str or tuple field identifier
        The weight field for computing weighted average for the profile
        values.  If None, the profile values are sums of the data in
        each bin.
    accumulation : bool or list of bools
        If True, the profile values for a bin n are the cumulative sum of
        all the values from bin 0 to n.  If -True, the sum is reversed so
        that the value for bin n is the cumulative sum from bin N (total bins)
        to n.  If the profile is 2D or 3D, a list of values can be given to
        control the summation in each dimension independently.
        Default: False.
    fractional : bool
        If True the profile values are divided by the sum of all
        the profile data such that the profile represents a probability
        distribution function.
    deposition : strings
        Controls the type of deposition used for ParticlePhasePlots.
        Valid choices are 'ngp' and 'cic'. Default is 'ngp'. This parameter is
        ignored the if the input fields are not of particle type.


    Examples
    --------

    Create a 1d profile.  Access bin field from profile.x and field
    data from profile[<field_name>].

    >>> ds = load("DD0046/DD0046")
    >>> ad = ds.all_data()
    >>> profile = create_profile(ad, [("gas", "density")],
    ...                              [("gas", "temperature"),
    ...                               ("gas", "velocity_x")])
    >>> print (profile.x)
    >>> print (profile["gas", "temperature"])

    """
    bin_fields = data_source._determine_fields(bin_fields)
    fields = ensure_list(fields)
    is_pfield = [data_source.ds._get_field_info(f).particle_type
                 for f in bin_fields + fields]
    wf = None
    if weight_field is not None:
        wf = data_source.ds._get_field_info(weight_field)
        is_pfield.append(wf.particle_type)
        wf = wf.name

    if len(bin_fields) > 1 and isinstance(accumulation, bool):
        accumulation = [accumulation for _ in range(len(bin_fields))]

    bin_fields = data_source._determine_fields(bin_fields)
    fields = data_source._determine_fields(fields)
    units = sanitize_field_tuple_keys(units, data_source)
    extrema = sanitize_field_tuple_keys(extrema, data_source)
    logs = sanitize_field_tuple_keys(logs, data_source)

    if any(is_pfield) and not all(is_pfield):
        raise YTIllDefinedProfile(
            bin_fields, data_source._determine_fields(fields), wf, is_pfield)
    elif len(bin_fields) == 1:
        cls = Profile1D
    elif len(bin_fields) == 2 and all(is_pfield):
        if deposition == 'cic':
            if logs is not None:
                if ((bin_fields[0] in logs and logs[bin_fields[0]]) or
                    (bin_fields[1] in logs and logs[bin_fields[1]])):
                    raise RuntimeError(
                        "CIC deposition is only implemented for linear-scaled "
                        "axes")
            else:
                logs = {bin_fields[0]: False, bin_fields[1]: False}
            if any(accumulation) or fractional:
                raise RuntimeError(
                    'The accumulation and fractional keyword arguments must be '
                    'False for CIC deposition')
        elif logs is None:
            logs = {bin_fields[0]: False, bin_fields[1]: False}
        cls = ParticleProfile
    elif len(bin_fields) == 2:
        cls = Profile2D
    elif len(bin_fields) == 3:
        cls = Profile3D
    else:
        raise NotImplementedError
    if weight_field is not None and cls == ParticleProfile:
        weight_field, = data_source._determine_fields([weight_field])
        if not data_source.ds._get_field_info(weight_field).particle_type:
            weight_field = None
    if not iterable(n_bins):
        n_bins = [n_bins] * len(bin_fields)
    if not iterable(accumulation):
        accumulation = [accumulation] * len(bin_fields)
    if logs is None:
        logs = {}
    logs_list = []
    for bin_field in bin_fields:
        if bin_field in logs:
            logs_list.append(logs[bin_field])
        else:
            logs_list.append(data_source.ds.field_info[bin_field].take_log)
    logs = logs_list
    if extrema is None:
        ex = [data_source.quantities["Extrema"](f, non_zero=l)
              for f, l in zip(bin_fields, logs)]
        # pad extrema by epsilon so cells at bin edges are not excluded
        for i, (mi, ma) in enumerate(ex):
            mi = mi - np.spacing(mi)
            ma = ma + np.spacing(ma)
            ex[i][0], ex[i][1] = mi, ma
    else:
        ex = []
        for bin_field in bin_fields:
            bf_units = data_source.ds.field_info[bin_field].output_units
            try:
                field_ex = list(extrema[bin_field[-1]])
            except KeyError:
                field_ex = list(extrema[bin_field])
            if isinstance(field_ex[0], tuple):
                field_ex = [data_source.ds.quan(*f) for f in field_ex]
            if any([exi is None for exi in field_ex]):
                try:
                    ds_extrema = data_source.quantities.extrema(bin_field)
                except AttributeError:
                    # ytdata profile datasets don't have data_source.quantities
                    bf_vals = data_source[bin_field]
                    ds_extrema = data_source.ds.arr(
                        [bf_vals.min(), bf_vals.max()])
                for i, exi in enumerate(field_ex):
                    if exi is None:
                        field_ex[i] = ds_extrema[i]
                        # pad extrema by epsilon so cells at bin edges are
                        # not excluded
                        field_ex[i] -= (-1)**i*np.spacing(field_ex[i])
            if units is not None and bin_field in units:
                for i, exi in enumerate(field_ex):
                    if hasattr(exi, 'units'):
                        field_ex[i] = exi.to(units[bin_field])
                    else:
                        field_ex[i] = data_source.ds.quan(exi, units[bin_field])
                fe = data_source.ds.arr(field_ex)
            else:
                if hasattr(field_ex, 'units'):
                    fe = field_ex.to(bf_units)
                else:
                    fe = data_source.ds.arr(field_ex, bf_units)
            fe.convert_to_units(bf_units)
            field_ex = [fe[0].v, fe[1].v]
            if iterable(field_ex[0]):
                field_ex[0] = data_source.ds.quan(field_ex[0][0], field_ex[0][1])
                field_ex[0] = field_ex[0].in_units(bf_units)
            if iterable(field_ex[1]):
                field_ex[1] = data_source.ds.quan(field_ex[1][0], field_ex[1][1])
                field_ex[1] = field_ex[1].in_units(bf_units)
            ex.append(field_ex)
    args = [data_source]
    for f, n, (mi, ma), l in zip(bin_fields, n_bins, ex, logs):
        args += [f, n, mi, ma, l]
    kwargs = dict(weight_field=weight_field)
    if cls is ParticleProfile:
        kwargs['deposition'] = deposition
    obj = cls(*args, **kwargs)
    setattr(obj, "accumulation", accumulation)
    setattr(obj, "fractional", fractional)
    if fields is not None:
        obj.add_fields([field for field in fields])
    for field in fields:
        if fractional:
            obj.field_data[field] /= obj.field_data[field].sum()
        for axis, acc in enumerate(accumulation):
            if not acc: continue
            temp = obj.field_data[field]
            temp = np.rollaxis(temp, axis)
            if weight_field is not None:
                temp_weight = obj.weight
                temp_weight = np.rollaxis(temp_weight, axis)
            if acc < 0:
                temp = temp[::-1]
                if weight_field is not None:
                    temp_weight = temp_weight[::-1]
            if weight_field is None:
                temp = temp.cumsum(axis=0)
            else:
                temp = (temp * temp_weight).cumsum(axis=0) / \
                  temp_weight.cumsum(axis=0)
            if acc < 0:
                temp = temp[::-1]
                if weight_field is not None:
                    temp_weight = temp_weight[::-1]
            temp = np.rollaxis(temp, axis)
            obj.field_data[field] = temp
            if weight_field is not None:
                temp_weight = np.rollaxis(temp_weight, axis)
                obj.weight = temp_weight
    if units is not None:
        for field, unit in units.items():
            field = data_source._determine_fields(field)[0]
            if field == obj.x_field:
                obj.set_x_unit(unit)
            elif field == getattr(obj, "y_field", None):
                obj.set_y_unit(unit)
            elif field == getattr(obj, "z_field", None):
                obj.set_z_unit(unit)
            else:
                obj.set_field_unit(field, unit)
    return obj



"""
Tests for making slices through raw fields

"""

#-----------------------------------------------------------------------------
# Copyright (c) 2017, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import yt
from yt.utilities.answer_testing.framework import \
    requires_ds, \
    data_dir_load, \
    GenericImageTest


def setup():
    """Test specific setup."""
    from yt.config import ytcfg
    ytcfg["yt", "__withintesting"] = "True"

def compare(ds, field, test_prefix, decimals=12):
    def slice_image(filename_prefix):
        sl = yt.SlicePlot(ds, 'z', field)
        sl.set_log('all', False)
        image_file = sl.save(filename_prefix)
        return image_file

    slice_image.__name__ = "slice_{}".format(test_prefix)
    test = GenericImageTest(ds, slice_image, decimals)
    test.prefix = test_prefix
    return test


raw_fields = "Laser/plt00015"
_raw_field_names =  [('raw', 'Bx'),
                     ('raw', 'By'),
                     ('raw', 'Bz'),
                     ('raw', 'Ex'),
                     ('raw', 'Ey'),
                     ('raw', 'Ez'),
                     ('raw', 'jx'),
                     ('raw', 'jy'),
                     ('raw', 'jz')]
@requires_ds(raw_fields)
def test_raw_field_slices():
    ds = data_dir_load(raw_fields)
    for field in _raw_field_names:
        yield compare(ds, field, "answers_raw_%s" % field[1])




"""
Unit test the ARMKDTree in yt.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.utilities.amr_kdtree.api import AMRKDTree
import yt.utilities.initial_conditions as ic
import yt.utilities.flagging_methods as fm
from yt.frontends.stream.api import load_uniform_grid, refine_amr
from yt.testing import assert_equal, assert_almost_equal, fake_amr_ds
import numpy as np
import itertools


def test_amr_kdtree_coverage():
    return #TESTDISABLED
    domain_dims = (32, 32, 32)
    data = np.zeros(domain_dims) + 0.25
    fo = [ic.CoredSphere(0.05, 0.3, [0.7, 0.4, 0.75],
                         {"density": (0.25, 100.0)})]
    rc = [fm.flagging_method_registry["overdensity"](8.0)]
    ug = load_uniform_grid({"density": data}, domain_dims, 1.0)
    ds = refine_amr(ug, rc, fo, 5)

    kd = AMRKDTree(ds)

    volume = kd.count_volume()
    assert_equal(volume, np.prod(ds.domain_right_edge - ds.domain_left_edge))

    cells = kd.count_cells()
    true_cells = ds.all_data().quantities['TotalQuantity']('Ones')[0]
    assert_equal(cells, true_cells)

    # This largely reproduces the AMRKDTree.tree.check_tree() functionality
    tree_ok = True
    for node in kd.tree.trunk.depth_traverse():
        if node.grid is None:
            continue
        grid = ds.index.grids[node.grid - kd._id_offset]
        dds = grid.dds
        gle = grid.LeftEdge
        nle = node.get_left_edge()
        nre = node.get_right_edge()
        li = np.rint((nle-gle)/dds).astype('int32')
        ri = np.rint((nre-gle)/dds).astype('int32')
        dims = (ri - li).astype('int32')
        tree_ok *= np.all(grid.LeftEdge <= nle)
        tree_ok *= np.all(grid.RightEdge >= nre)
        tree_ok *= np.all(dims > 0)

    assert_equal(True, tree_ok)

def test_amr_kdtree_set_fields():
    ds = fake_amr_ds(fields=["density", "pressure"])
    dd = ds.all_data()

    fields = ds.field_list
    dd.tiles.set_fields(fields, [True, True], False)
    gold = {}
    for i, block in enumerate(dd.tiles.traverse()):
        gold[i] = [data.copy() for data in block.my_data]

    for log_fields in itertools.product([True, False], [True, False]):
        dd.tiles.set_fields(fields, log_fields, False)
        for iblock, block in enumerate(dd.tiles.traverse()):
            for i in range(len(fields)):
                if log_fields[i]:
                    data = block.my_data[i]
                else:
                    data = np.log10(block.my_data[i])
                assert_almost_equal(gold[iblock][i], data)



"""
ExodusII-specific fields



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.fields.field_info_container import \
    FieldInfoContainer

# We need to specify which fields we might have in our dataset.  The field info
# container subclass here will define which fields it knows about.  There are
# optionally methods on it that get called which can be subclassed.

class ExodusIIFieldInfo(FieldInfoContainer):
    known_other_fields = (
        # Each entry here is of the form
        # ( "name", ("units", ["fields", "to", "alias"], # "display_name")),
    )

    known_particle_fields = (
        # Identical form to above
        # ( "name", ("units", ["fields", "to", "alias"], # "display_name")),
    )

    def __init__(self, ds, field_list):
        super(ExodusIIFieldInfo, self).__init__(ds, field_list)
        for name in self:
            self[name].take_log = False
        # If you want, you can check self.field_list

    def setup_fluid_fields(self):
        # Here we do anything that might need info about the dataset.
        # You can use self.alias, self.add_output_field and self.add_field .
        pass

    def setup_particle_fields(self, ptype):
        # This will get called for every particle type.
        pass



"""
GDF-specific fields



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.fields.field_info_container import \
    FieldInfoContainer

# The nice thing about GDF is that for the most part, everything is in CGS,
# with potentially a scalar modification.

class GDFFieldInfo(FieldInfoContainer):
    known_other_fields = (
        ("density", ("g/cm**3", ["density"], None)),
        ("specific_energy", ("erg/g", ["thermal_energy"], None)),
        ("pressure", ("erg/cm**3", ["pressure"], None)),
        ("temperature", ("K", ["temperature"], None)),
        ("velocity_x", ("cm/s", ["velocity_x"], None)),
        ("velocity_y", ("cm/s", ["velocity_y"], None)),
        ("velocity_z", ("cm/s", ["velocity_z"], None)),
        ("mag_field_x", ("gauss", ["magnetic_field_x"], None)),
        ("mag_field_y", ("gauss", ["magnetic_field_y"], None)),
        ("mag_field_z", ("gauss", ["magnetic_field_z"], None)),
    )
    known_particle_fields = ()

    def setup_fluid_fields(self):
        from yt.fields.magnetic_field import \
            setup_magnetic_field_aliases
        setup_magnetic_field_aliases(self, "gdf", ["magnetic_field_%s" % ax for ax in "xyz"])



"""
A mechanism for listing available analysis modules.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import os
import sys

def get_available_modules():
    modpath = os.path.abspath(os.path.dirname(__file__))
    available_modules = []
    for d in [os.path.join(modpath, f) for f in os.listdir(modpath)]:
        if os.path.isdir(d) and os.path.isfile(os.path.join(d, "api.py")):
            available_modules.append(os.path.basename(d))
    return available_modules

class AnalysisModuleLoader(object):

    @property
    def available_modules(self):
        return get_available_modules()

    def __getattr__(self, attr):
        try:
            name = "yt.analysis_modules.%s.api" % (attr)
            __import__(name, level=-1)
            setattr(self, attr, sys.modules[name])
        except ImportError:
            raise AttributeError(attr)
        return getattr(self, attr)

    def __dir__(self):
        # This is a badly behaving object.  I was unable to get this line:
        #return super(AnalysisModuleLoader, self).__dir__() + self.available_modules
        # to work, so we simply return only the methods we know about.
        return ["available_modules"] + self.available_modules

amods = AnalysisModuleLoader()
